# Investments::Journal.pm - The Object Class that provides a Investments::Journal Object
#$Id: Journal.pm,v 1.38.6.2 2006/07/12 20:54:24 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.

package Portal::Accounting::Investments::Journal;
use strict;
use Portal::Accounting::Investments::Main;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Investments::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.38.6.2 $';

# global variables.

# modify to represent the commands this state provides.
my %commands = (
  display => "Investment Journal",
  new => "New Investment",
  edit => "Edit Investment",
  delete => "Delete Investment",
  save => "Save Investment Transaction",
  searchForm => "Stock Transactions",
  search => "Stock Transactions",
);

=head1 NAME

Investments::Journal - Object used to work with Investments::Journal

=head1 SYNOPSIS

  use Portal::Accounting::Investments::Journal;
  my $obj = Portal::Accounting::Investments::Journal->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

Investments::Journal is the Investments::Journal Menu from which every other part of
the systems will be accessed.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(codeVersion, release, DBVersion)

 Creates a new instance of the Portal::Accounting::Investments::Journal module.
 See Portal::Accounting::Base(3) for a listing of required arguments.

 Instantiates an instance of the Investments::Journal State. !!Anything passed in here should be
 data that never changes from one run of a command to another such as database handles
 portalSession? input? cookieObj? dont fit that criterion

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{commands} = \%commands;

  # do validation
  if (!$self->Portal::Accounting::Investments::Journal::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  return $self;
}

=item % getCommands(void)

  This method returns the commands hash and is only to be
  used by the registerApp method for creating the permissions,
  if using the AppStateCommandSecurity implementation.

=cut
sub getCommands
{
  return %commands;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item HTMLObject c_display()

=cut
sub c_display
{
  my $self = shift;
  my %args = ( command => "display", @_ );
  my $command = $args{command};
  my $doc = $self->prepDoc;
  my $l = $self->langObj;
  my $lc;
  my $Buy = $l->map("buy");
  my $Sell = $l->map("sell");
  my $Fee = $l->map("fee");
  my $Memo = $l->map("memo");
  my $Merger = $l->map("acqs");
  my $Return = $l->map("rtcp");
  my $Spin = $l->map("spof");
  my $Split = $l->map("splt");
  my $StockTx = $l->map("sxfr");
  use Portal::Accounting::Investments::Objects::Search;
  my $search = Portal::Accounting::Investments::Objects::Search->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
  my %form = $search->printForm;
  if ($search->error) { $doc->print($search->errorMessage); }
  else { $doc->print(%form); }
  $doc->print(<<"FOO");
<p>
  Investment Transactions:<br>
  <a href="?$self->{getInfo}&command=new&type=buy">$Buy</a><br>
  <a href="?$self->{getInfo}&command=new&type=sell">$Sell</a><br>
  <a href="?$self->{getInfo}&command=new&type=fee">$Fee</a><br>
  <a href="?$self->{getInfo}&command=new&type=memo">$Memo</a><br>
  <a href="?$self->{getInfo}&command=new&type=acqs">$Merger</a><br>
  <a href="?$self->{getInfo}&command=new&type=rtcp">$Return</a><br>
  <a href="?$self->{getInfo}&command=new&type=spof">$Spin</a><br>
  <a href="?$self->{getInfo}&command=new&type=splt">$Split</a><br>
  <a href="?$self->{getInfo}&command=new&type=sxfr">$StockTx</a><br>
</p>
FOO
  return $doc;
}

sub c_edit
{
  my $self = shift;
  if (exists $self->{objects}->{objects}->{$self->{input}->{type}})
  {
     eval('use Portal::Accounting::Investments::Objects::'.$self->{objects}->{objects}->{$self->{input}->{type}});
     eval('$self->{OBJECT} = Portal::Accounting::Investments::Objects::'.$self->{objects}->{objects}->{$self->{input}->{type}}.'->new(dbHandle => $self->{dbHandle}, ledgerCode => $self->{input}->{ledgerCode}, langObj => $self->{langObj}, input => $self->{input});');
  }
  else
  {
    my $doc = $self->prepDoc;
    $doc->setFocus("body");
    $doc->print("<p>Invalid type sent to Investments::Journal->c_new. '$self->{input}->{type}'</p>\n");
    return $doc;
  }
  my $doc = $self->SUPER::c_edit(@_);
#  $doc->print($self->mainButton);
  return $doc;
}

sub c_save
{
  my $self = shift;
  my $doc = $self->SUPER::c_save(@_);
  my $message = '<div>SUCCESS</div>';
  if (!$self->error && defined $self->{OBJECT})
  {
    $message .= $self->{OBJECT}->printHTML;
    $doc->print($self->sameButton($message). $message);
  }
  return $doc;
}

sub prepDoc
{
  my $self = shift;
  my $doc = $self->SUPER::prepDoc();
  $doc->setJavascriptInclude(file => $self->{jsDir}."Investments.js");
  return $doc;
}

sub c_searchForm
{
  my $self = shift;
  my $doc = $self->prepDoc;
  my %args = ( command => $self->{input}->{command}, @_ );
  my $command = $args{command};
  my $l = $self->{langObj};
  use Portal::Accounting::Investments::Objects::Search;
  my $foo = Portal::Accounting::Investments::Objects::Search->new(langObj => $self->{langObj}, dbHandle => $self->{dbHandle}, ledgerCode => $self->{input}->{ledgerCode}, @_);
  if ($foo->error) { $doc->print($foo->errorMessage); return $doc;  }
  $self->{OBJECT} = $foo;
  if ($command eq "search")
  { #validate user input then react
    if ($foo->populate(%{$self->{input}}) && $foo->validateForm(input => $self->{input}))
    {
      my @je = $foo->selectArray();
#      $doc->print( (scalar @je) . $foo->debugString);
#      $doc->print( $foo->{LIST}[1]);
      if ($foo->error) { $doc->print($foo->errorMessage); }
      else
      {
       $doc->print("<div id=\"main\">\n");
       # then display them by walking and $o->printHTML
        $doc->print("<table width=\"100%\"><tr valign=\"top\"><td><b>"
        .$l->map('Account')."</b> ".$foo->lookupSelectName(item => 'accounts')
        ."</td><td align=\"left\">".$foo->lookupSelectName(item => 'stocks')."</td>\n<td align=\"right\">");
        if (scalar @je == 0) { $doc->print($l->map("Found")." 0</td></tr></table>\n"); }
        else
        {
          #output search paging form
          my %form = $foo->pagerForm(number => $self->{input}->{number});
          if ($foo->error) { $doc->print($foo->errorMessage); }
          else { $doc->print(%form); }
          $doc->print("</td></tr></table>\n");
          $doc->print("<form name=\"main\">\n");
          my $count = 0;
          my $historical = '';
    foreach my $bar (@je)
          {
            if ($count++ == 0)
            {
              $doc->print($je[0]->printHeaders);
              $historical = $je[0]->{ledger}->preOpenBooksDate; #initialize historical flag
              if ($bar->get('date') > $historical) { $historical = ''; }  #reset if already past historical date
            }
            if ($historical ne '' && $bar->get('date') > $historical)  #if flag is set and curent transaction is passed historical date.
            {
              $doc->print('<div class="tertiary"><small>All transactions above occurred before the books opened.</small></div>');
              $historical = '';  #turn flag off
            }
            $bar->isValid;
            $doc->print($bar->printReport);
            $doc->print($bar->errorMessage);
            $doc->print($bar->debugString);
          }
          if ($self->warning) { $doc->print($self->warningString); }
          $doc->print($self->debugString);
          $doc->print(<<"FOO");
    <input type="hidden" name="ledgerCode" value="$self->{input}->{ledgerCode}" />
  </form>
</div>
FOO
          return $doc;
        }
      }
    }
  }
  #display form
  my %form = $foo->printForm();
  if ($foo->error)
  {
    $self->error($foo->errorMessage);
  }
  if (defined %form)
  {
    $doc->print(%form);
  }
  else
  {
    $doc->print('error! form is undefined<br>');
    $doc->print("\nformObj:".$foo->{formObj}->error);
  }

  $doc->print($foo->debugString);
  return $doc;
}

sub c_search
{
  my $self = shift;
  return $self->c_searchForm(@_);
}

sub c_delete
{
  my $self = shift;
  use JSON;
  my %obj;
  my $doc = HTMLObject::Normal->new;
  $doc->setContentType('text/plain');
  #call backend code
  my $event = Portal::Accounting::Investments::Objects::Event->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $self->{input}->{ledgerCode});
  if ($event->error) { $obj{ERROR} = $event->errorMessage; }
  my $event2 = $event->getEvent($self->{input}->{id});
  if ($event->error) { $obj{ERROR} = $event->errorMessage; }
  elsif ($event2->error) { $obj{ERROR} = $event2->errorMessage; }
  else
  {
    my $success = $event2->deleteEvents();
    if (!$event2->error)
    {
      $obj{ERROR} = $self->{langObj}->map('Success');
      $obj{STATUS} = $self->{input}->{id};
#      $obj{nodes} = $event2->{nodes};  #does not work because cant easily get a list of deleted nodes
    }
    else
    {
      $obj{ERROR} = '<div class="error">'.$self->{langObj}->map('Failed') . "</div>".$event2->errorMessage;
      $obj{STATUS} = -1;
    }
  }
  $doc->print("//<!DOCTYPE JSON>\n");
  $doc->print("var obj = (".objToJson(\%obj).");");
  return $doc;
}

sub sameButton
{
  my $self = shift;
  my $message = shift;
  if (!defined $message) { $message = ''; }
  my $r = <<"FOO";
<form name="display">
  <input type="hidden" name="app" value="$self->{app}" />
  <input type="hidden" name="state" value="Investments::Journal" />
  <input type="hidden" name="command" value="new" />
  <input type="hidden" name="ledgerCode" value="$self->{input}->{ledgerCode}" />
  <input type="hidden" name="type" value="$self->{input}->{type}" />
  <input type="hidden" name="date" value="$self->{input}->{date}" />
  <input type="hidden" name="account" value="$self->{input}->{account}" />
  <input type="hidden" name="account2" value="$self->{input}->{account2}" />
  <input type="hidden" name="stock" value="$self->{input}->{stock}" />
  <input type="submit" value="Enter Another" />
  <input type="hidden" name="MESSAGE" value="$message" />
</form>
FOO
  return $r;
}

=back

=cut

1;
__END__

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::BaseL(3)

=cut
