# Investment.pm - The Object Class that provides a Investment Object
#$Id: Base.pm,v 1.12 2006/01/05 22:35:13 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2003 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Investments::Objects::Base;
use strict;
use Portal::Accounting::Objects::BaseL;
use Portal::Accounting::Objects::Account;
use Portal::Accounting::Objects::SubAccount;
use Portal::Accounting::Objects::SubEntity;
use Portal::Accounting::Investments::Objects::Stock;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT %Elements);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.12 $';

=head1 NAME

Investments::Objects::Base - Object for deriving other investment objects.

=head1 SYNOPSIS

  use Portal::Accounting::Investments::Objects::Base;
  @ISA = qw(Portal::Accounting::Investments::Objects::Base Exporter AutoLoader);

=head1 DESCRIPTION

Investments::Object::Base is a base class for investment objects.  It holds common data and methods to be used by the investment objects.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(langObj =>, dbHandle =>, ledgerCode => )

    Creates a new instance of Portal::Accounting::Investments::Objects::Base.
    See Portal::Objects::Base(3) if required arguments listed above don't work properly.

  Defines these variables:
  
  $self->{objects} = {        #old numbering system--deprecated
    buy  => "Buy",            #10
    sell => "Sell",           #20
    adjs  => "Adjustment",     #40 used to bee fee
    memo => "Memo",           #memo
    acqs => "Merger",         #80
    rtcp => "ReturnOfCapital", #50
    spof => "SpinOff",        #70, #60 was stock swap. deprecated.
    splt => "StockSplit",     #90
    sdiv => "StockDividend",  #100  #deprecated!!! will remove in the future
    sxfr => "StockTransfer"   #30
  };


=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (ledgerCode => '', @_);

  if ($self->error)
  {
    return $self;
  }

  # instantiate anything unique to this module
  $self->{objects} = {
    buy  => "Buy",            #10
    sell => "Sell",           #20
    adjs  => "Adjustment",     #40
    memo => "Memo",           #memo
    acqs => "Merger",         #80
    rtcp => "ReturnOfCapital", #50
    spof => "SpinOff",        #70, #60 was stock swap. deprecated.
    splt => "StockSplit",     #90
    sdiv => "StockDividend",  #100  #deprecated!!! will remove in the future
    sxfr => "StockTransfer"   #30
  };

  $self->{shortTermCG} = 4200;
  $self->{longTermCG} = 4201;
  $self->{shortTermCL} = 5200;
  $self->{longTermCL} = 5201;
  $self->{Journal} = "IS";
  
  # do anything else you might need to do.

  my $chart = Portal::Accounting::Objects::Account->new(langObj => $self->langObj, dbHandle => $self->{dbHandle}, ledgerCode => $args{ledgerCode});
  $chart->set('type', 3);
  $self->{accounts} = $chart->selectOptionsByType();
  if ($chart->error) { $self->error("Failed to get accounts<br>".$chart->errorMessage."\n"); return $self;  }
  if (! defined $self->{accounts}) { $self->{accounts} = {names => [], values => [] }; }

  my $stock = Portal::Accounting::Investments::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->{dbHandle});
  $self->{stocks} = $stock->selectOptions();
  if ($stock->error) { $self->error("Failed to stocks<br>".$stock->errorMessage."\n"); return $self;  }
  if (! defined $self->{stocks}) { $self->{stocks} = {names => [], values => [] }; }
  
#   my $asset = Portal::Accounting::Objects::asset->new(langObj => $self->langObj, dbHandle => $self->{dbHandle});
#   $self->{assets} = $asset->selectOptions();
  $self->{assets} = {names => [], values => []};
#   if ($asset->error) { $self->error("Failed to assets<br>".$asset->errorMessage."\n"); return $self;  }
#   if (! defined $self->{assets}) { $self->{assets} = {names => [], values => [] }; }

  my $subE = Portal::Accounting::Objects::SubEntity->new(langObj => $self->langObj, dbHandle => $self->{dbHandle}, ledgerCode => $args{ledgerCode});
  $self->{subEntities} = $subE->selectOptions();
  if ($subE->error) { $self->error("Failed to stocks<br>".$subE->errorMessage."\n"); return $self;  }
  if (! defined $self->{subEntities}) { $self->{subEntities} = {names => [], values => [] }; }

  return $self;
}

=item bool isValid(void)

 Returns 0 for invalid or 1 for valid.
 The error will be available via errorMessage().

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub isNew { return 1; }

sub printReport
{
  my $self = shift;
  return $self->SUPER::printForm(@_, form => 'report');
}
=back

=cut

1;
__END__

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Objects::Base(3)

=cut