# Portfolio.pm - The Object Class that provides a Portfolio Object
#$Id: Portfolio.pm,v 1.20.4.1 2006/08/15 19:38:13 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2005 http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Investments::Objects::Portfolio;
use strict;
use Portal::Accounting::Investments::Objects::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Investments::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.20.4.1 $';

=head1 NAME

Portfolio - Object used to create, edit, save an Portfolio.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Portfolio;
  my $obj = Portal::Accounting::Objects::Portfolio->new();
  $obj->populate( );
  if ($obj->error)
  {
    die $errorMessage();
  }

=head1 DESCRIPTION

Portfolio is a Portfolio class.  This represents a sub account in the
accounting system.  Right now number is used internally. Name and code
are user defined.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(dbHandle => , langObj =>)

 Creates a new instance of the Portal::Accounting::Objects::Portfolio module.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( ledgerCode => "", app => "Accounting", state => "Investments::Portfolio", @_ );

  # instantiate anything unique to this module
  $self->{dbHandle} = $args{dbHandle};
  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }
  # instantiate anything unique to this module
  my $me = "Portfolio";
  my $l = $self->{langObj};
  my @date = $self->getDateArray;
  my $today = "$date[0]-$date[1]-$date[2]";

  my $collapse = $self->{imageDir}.'/minus.gif';
  my $expand = $self->{imageDir}.'/plus.gif';
  $self->{data}->{main} = {
    "onload" => { -Type => "hidden", },
    "app" => { -Type => "hidden", -Value => $args{app} },
    "state" => { -Type => "hidden", -Value => $args{state} },
    "command" => { -Type => "hidden", -Value => "search" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, -DataType => "string" },
    "today" => { -Type => "hidden", -Value => $today, },
    "date" => { -Label => $l->map("Date"), -Type => "date-picker", -DataType => "string",
      -year => $date[0], -Value => $today },
    "account" => { -Label => "#LANG=Account#", -Type => "multi-select", -DataType => "int",
      -Options => $self->{accounts}, size => 4,
      -Buttons => [ "SelectAll", "ToggleSelect" ],
    },
    "stock" => { -Label => $l->map("Stock"), -Type => "multi-select", -DataType => "int",
      -Options => $self->{stocks}, size => 4,
      -Buttons => [ "SelectAll", "ToggleSelect" ],
    },
    "format" => { -Label => $l->map("Format"), -Type => "select", -DataType => "int",
      -Options => { names => [ 'html', 'csv' ], values => [qw(html csv)] }, -Value => 'html',  },
  };
  $self->{data}->{report} = {
    "onload" => { -Type => "hidden", -onload => "var images = new Array(); images['field'] = 'label'; images['expand'] = '$expand'; images['collapse'] = '$collapse'; walkRows(document.getElementById('portfolio'), '', images);", },
    "date" => { -Label => $l->map("Date"), -Type => "date-picker", -DataType => "string",
      -year => $date[0], -Value => $today, -class => 'date', class => 'date',
    },
    "app" => { -Type => "hidden", -Value => $args{app} },
    "state" => { -Type => "hidden", -Value => $args{state} },
    "command" => { -Type => "hidden", -Value => "search" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, -DataType => "string" },
    "account" => { -Type => "hidden", -DataType => "integer"   },
    "stock" => { -Type => "hidden", -DataType => "integer" ,  },
    "format" => { -Label => $l->map("Format"), -Type => "select", -DataType => "int",
      -Options => { names => [ 'html', 'csv' ], values => [qw(html csv)] }, -Value => 'html', },
 };
  $self->{css}->{main} = <<"ENDOF";
ENDOF
  $self->{css}->{report} = <<"ENDOF";
  table.portfolio { width: 100%; }
  th.level { display: none; }
  th.label { text-align: left; }
  th.account_num { display: none; }
  th.account { text-align: left; display: none; }
  th.sub_account { display: none; }
  th.stock { text-align: left; display: none; }
  th.lot { text-align: left; display: none; }
  th.shares { text-align: right; }
  th.cost { text-align: right; }
  th.cost_share { text-align: right; }
  th.price { text-align: right; }
  th.mkt { text-align: right; }
  th.gain_loss { text-align: right; }
  th.ticker { display: none; }
  th.exchange { display: none; }
  td.level { display: none; }
  td.label { text-align: left; }
  td.account_num { display: none; }
  td.account { text-align: left; display: none; }
  td.sub_account { display: none; }
  td.stock { text-align: left; display: none; }
  td.lot { text-align: left; display: none; }
  td.shares { text-align: right; }
  td.cost { text-align: right; }
  td.cost_share { text-align: right; }
  td.price { text-align: right; }
  td.mkt { text-align: right; }
  td.gain_loss { text-align: right; }
  td.ticker { display: none; }
  td.exchange { display: none; }
  text.date { border: none; border-bottom-style: solid; }
  .level1 { font-size: 1.2em; }
  .level2 { font-size: 1.1em; padding-left: 2em; }
  .level3 { font-size: 1em; padding-left: 4em;}
  .level4 { font-size: .9em; padding-left: 8em;}
ENDOF
  $self->{template}->{main} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIELD=app#
  #FIELD=state#
  #FIELD=command#
  #FIELD=ledgerCode#
  #FIELD=today#
  <table class="primary">
    <tr>
      <td><b>#LABEL=date#</b> #FI=date#</td>
      <td><b>#LABEL=format# #FI=format#</td>
    </tr>
    <tr>
      <td><b>#LABEL=account#</b><br>#IF=account#</td>
      <td><b>#LABEL=stock#</b><br>#IF=stock#</td>
    </tr>
  </table>
  <input type="submit" name="submit" value="#LANG=Submit#">
</form>
ENDOF
  $self->{template}->{report} = <<"ENDOF";
<div style="text-align: center;">
  <form name="report">
    #FIELD=app#
    #FIELD=state#
    #FIELD=command#
    #FIELD=ledgerCode#
    #FIELD=account#
    #FIELD=stock#
    #FIELD=date#
    <span class="printHide">
      <input type="submit" name="submit" value="*" title="Fetch Data For New Date" />
      &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
      <span onclick="tableShow(document.getElementById('portfolio'), document.getElementById('level').value);">Show</span>
      <select id="level" name="level">
        <option value="2">Accounts</option>
        <option value="3">Stocks</option>
        <option value="4">Lots</option>
      </select>
      <span onclick="tableHide(document.getElementById('portfolio'), document.getElementById('level').value);">Hide</span>
    </span>

  </form>
</div>
ENDOF
  $self->{javascript}->{main} = <<"ENDOF";
ENDOF
  $self->{profile}->{main} = {
    required => [ qw( date ) ],
    optional => [ qw( format stock account ) ],
    constraints => {
         date => qr/^\d{4}-\d{1,2}-\d{1,2}$/,
         stock => qr/^\d+$/,
         format => qr/^csv|html$/,
         account => qr/^\d+$/,
       },
  };
  $self->{profile}->{report} = {
    required => [ qw( date) ],
    optional => [  ],
    constraints => {
         date => qr/^\d{4}-\d{1,2}-\d{1,2}$/,
         stock => qr/^\d+$/,
         format => qr/^csv|html$/,
         account => qr/^\d+$/,
       },
  };
  $self->{columns} = [ qw(label account_num account sub_account stock lot purchased shares cost cost_share price mkt gain_loss ticker exchange) ];
  $self->{columnNames} = {
    label => 'Account/Stock/Lot',
    level => 'level',
    account_num => 'account_num',
    account => 'Account',
    sub_account => 'sub_account',
    stock => 'Stock',
    lot => 'Lot',
    purchased => 'Purchased',
    shares => 'Shares',
    cost => 'Cost',
    cost_share => 'Cost/Share',
    price => 'Price',
    mkt => 'Mkt Value',
    gain_loss => 'Gain/Loss',
    ticker => 'ticker',
    exchange => 'exchange',
  };

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item boolean select

  This routine gathers the xml data for this portfolio and subtotals it.

=cut

=item selectSQL

routine to generate sql neeeded to select data

=cut
sub selectSQL
{
  my $self = shift;
  my $date = $self->get('date');
  my $subAccountNum = $self->get('stock');
  my $ledgerCode = $self->get('ledgerCode');
  my $accountNum = $self->get('account');
  my $aSQL = '';
  my $sSQL = '';
  my $Total = $self->{langObj}->map('Total');
  if ($subAccountNum ne "")
  {
    my @a = split('\0', $subAccountNum);
    if (scalar @a != scalar @{$self->{stocks}->{values}})
    {
      $sSQL = "     AND sub_account_num IN (".join(",",@a).")";
    }
  }
  if ($accountNum ne "")
  {
    my @a = split('\0', $accountNum);
    $aSQL = "     AND account_num IN (".join(",",@a).")";
  }

  #cant format shares with commas bc it trashes calculations later
  $self->{SELECT}[0] = <<"FOO";
  SELECT '4' as level,
    '<span class="level4">'||l.lot||'</span>' AS label,
    l.account_num::text,
    a.name AS account,
    l.sub_account_num::text AS sub_account,
    s.sub_acc_name AS stock,
    l.lot::text,
    (SELECT MIN(buy_date) from sec_event_tb AS sec WHERE sec.l = l.buy_lot AND sec.buy_date IS NOT NULL)::text AS purchased,
    CASE WHEN l.shares - trunc(l.shares) < 0.0000001 THEN (SELECT trunc(l.shares)::text ) ELSE (SELECT l.shares::text) END AS shares,
    l.cost::text AS cost,
    to_char(l.cost/l.shares, '9G999G999D00') AS cost_share,
    price_wrapper('$date', l.sub_account_num)::text AS price,
    price_wrapper('$date', l.sub_account_num) * l.shares AS mkt,
    price_wrapper('$date', l.sub_account_num) * l.shares - l.cost AS gain_loss,
    stocks_tb.ticker AS ticker,
    stocks_tb.market AS exchange
  FROM lot_node_tb AS l, account_tb AS a, sub_acc_tb AS s, stocks_tb
  WHERE
    ledger_code = '$ledgerCode'
    AND l.account_num = a.number
    AND l.sub_account_num = s.sub_acc_number
    AND l.sub_account_num = stocks_tb.sub_acct
    AND l.shares > 0
    AND l.event_date <= '$date'
    AND l.node_id IN
    (
      SELECT max(lot_node_tb.node_id) AS id
      FROM lot_node_tb, sec_event_tb
      WHERE
        lot_node_tb.event_ref = sec_event_tb.event_ref
        AND (
          lot_node_tb.event_date < '$date'
          OR (
            lot_node_tb.event_date = '$date'
            AND sec_event_tb.counter < 1
          )
        )
      GROUP BY lot
    )
    $aSQL
    $sSQL

  UNION ALL
  SELECT '3' as level,
    '<span class="level3">'||s.sub_acc_name||'</span>' AS label,
    l.account_num::text,
    a.name AS account,
    l.sub_account_num::text AS sub_account,
    s.sub_acc_name AS stock,
    '' AS lot,
    '' AS purchased,
    CASE
      WHEN SUM(l.shares) - trunc(SUM(l.shares)) < 0.0000001 THEN
        (SELECT trunc(SUM(l.shares))::text )
      ELSE
        (SELECT SUM(l.shares)::text )
    END AS shares,
    SUM(l.cost)::text AS cost,
    to_char( SUM(l.cost) / SUM(l.shares), '999G999G999D00')::text AS cost_share,
    price_wrapper('$date', l.sub_account_num)::text AS price,
    SUM(price_wrapper('$date', l.sub_account_num) * l.shares) AS mkt,
    SUM(price_wrapper('$date', l.sub_account_num) * l.shares - l.cost) AS gain_loss,
    stocks_tb.ticker AS ticker,
    stocks_tb.market AS exchange
  FROM lot_node_tb AS l, account_tb AS a, sub_acc_tb AS s, stocks_tb
  WHERE
    ledger_code = '$ledgerCode'
    AND l.account_num = a.number
    AND l.sub_account_num = s.sub_acc_number
    AND l.sub_account_num = stocks_tb.sub_acct
    AND l.shares > 0
    AND l.event_date <= '$date'
    AND l.node_id IN
    (
      SELECT max(lot_node_tb.node_id) AS id
      FROM lot_node_tb, sec_event_tb
      WHERE
        lot_node_tb.event_ref = sec_event_tb.event_ref
        AND (
          lot_node_tb.event_date < '$date'
          OR (
            lot_node_tb.event_date = '$date'
            AND sec_event_tb.counter < 1
          )
        )
      GROUP BY lot
    )
    $aSQL
    $sSQL
  GROUP BY l.account_num, l.sub_account_num, a.name, s.sub_acc_name, stocks_tb.market, stocks_tb.ticker

  UNION ALL
  SELECT '2' as level,
    '<span class="level2">'||a.name||'</span>' AS label,
    l.account_num::text,
    a.name AS account,
    '' AS sub_account,
    '' AS stock,
    '' AS lot,
    '' AS purchased,
    '' AS shares,
    SUM(l.cost)::text AS cost,
    '' AS cost_share,
    '' AS price,
    SUM(price_wrapper('$date', l.sub_account_num) * l.shares) AS mkt,
    SUM(price_wrapper('$date', l.sub_account_num) * l.shares - l.cost) AS gain_loss,
    '' AS ticker,
    '' AS exchange
  FROM lot_node_tb AS l, account_tb AS a, sub_acc_tb AS s
  WHERE
    ledger_code = '$ledgerCode'
    AND l.account_num = a.number
    AND l.sub_account_num = s.sub_acc_number
    AND l.sub_account_num = stocks_tb.sub_acct
    AND l.shares > 0
    AND l.event_date <= '$date'
    AND l.node_id IN
    (
      SELECT max(lot_node_tb.node_id) AS id
      FROM lot_node_tb, sec_event_tb
      WHERE
        lot_node_tb.event_ref = sec_event_tb.event_ref
        AND (
          lot_node_tb.event_date < '$date'
          OR (
            lot_node_tb.event_date = '$date'
            AND sec_event_tb.counter < 1
          )
        )
      GROUP BY lot
    )
    $aSQL
    $sSQL
  GROUP BY l.account_num, a.name

  UNION ALL
  SELECT '1' as level,
    '<span class="level1">$Total</span>' AS label,
    '' AS account_num,
    '' AS account,
    '' AS sub_account,
    '' AS stock,
    '' AS lot,
    '' AS purchased,
    '' AS shares,
    SUM(l.cost)::text AS cost,
    '' AS cost_share,
    '' AS price,
    SUM(price_wrapper('$date', l.sub_account_num) * l.shares) AS mkt,
    SUM(price_wrapper('$date', l.sub_account_num) * l.shares - l.cost) AS gain_loss,
    '' AS ticker,
    '' AS exchange
  FROM lot_node_tb AS l, account_tb AS a, sub_acc_tb AS s
  WHERE
    ledger_code = '$ledgerCode'
    AND l.account_num = a.number
    AND l.sub_account_num = s.sub_acc_number
    AND l.sub_account_num = stocks_tb.sub_acct
    AND l.shares > 0
    AND l.event_date <= '$date'
    AND l.node_id IN
    (
      SELECT max(lot_node_tb.node_id) AS id
      FROM lot_node_tb, sec_event_tb
      WHERE
        lot_node_tb.event_ref = sec_event_tb.event_ref
        AND (
          lot_node_tb.event_date < '$date'
          OR (
            lot_node_tb.event_date = '$date'
            AND sec_event_tb.counter < 1
          )
        )
      GROUP BY lot
    )
    $aSQL
    $sSQL
  GROUP BY label

  ORDER BY account, stock, level ASC
FOO
  $self->{SELECT}[1] = <<"FOO";
  SELECT '1' as level,
    NULL AS account_num,
    '' AS account,
    NULL AS sub_account,
    '' AS stock,
    NULL AS lot,
    NULL AS shares,
    SUM(l.cost),
    NULL AS cost_share,
    NULL AS price,
    to_char( SUM(price_wrapper('$date', l.sub_account_num) * l.shares) , '9G999G999D00') AS mkt,
    to_char( SUM(price_wrapper('$date', l.sub_account_num) * l.shares - l.cost), '9G999G999D00') AS gain_loss,
    NULL AS ticker,
    NULL AS exchange
  FROM lot_node_tb AS l, account_tb AS a, sub_acc_tb AS s
  WHERE node_id IN
  (
    SELECT max (node_id)
    FROM lot_node_tb AS b
    WHERE ledger_code = '$ledgerCode' AND shares > 0
      AND event_date = ( SELECT max(event_date) FROM lot_node_tb WHERE lot = b.lot AND event_date <= '$date' )
    GROUP BY lot
  )
    AND l.account_num = a.number
    AND l.sub_account_num = s.sub_acc_number
    AND l.sub_account_num = stocks_tb.sub_acct
    $aSQL
    $sSQL
  ORDER BY account_num, sub_account, level ASC
FOO
  return @{$self->{SELECT}};
}

=item string printXML

  this routine outputs the xml document represented by this object.

=cut
sub printXML
{
  my $self = shift;
  my @sql = $self->selectSQL;
  my $xml = $self->{dbHandle}->readXML(sql => $sql[0], columns => 2);
  if ($self->dbHandle->error) { $self->error($self->errorMessage); return ''; }
  return $xml;
}

sub printReport
{
  my $self = shift;
  my $message = '';
  my @sql = $self->selectSQL;
  my %r = $self->SUPER::printForm(@_, form => 'report');
  $r{css} .= $self->{css}->{report};
  my $format = $self->get('format');
  my $date = $self->get('date');
  if ($format eq 'csv') { $r{body} = "'$date'\n";}
  my $table;
  my %quotes;

  if ($self->error) {
    $r{body} .= $self->{formObj}->genErrorString('all');
    return %r;
  }
  use Date::Calc qw(Today);
  my @today = Today;
  my $today = $today[0].sprintf("%02d", $today[1]).sprintf("%02d", $today[2]);
  if ($date =~ /(\d{4})-(\d{1,2})-(\d{1,2})/)
  {
    $date = $1.sprintf("%02d", $2).sprintf("%02d", $3);
  }
  my $start = time;
  my @data = $self->{dbHandle}->getDataHash(sql => $sql[0] );
  if ($self->debug) { $self->debug('<div>Time to gather Data: '. (time - $start) .' seconds</div>'); }
  if ($self->dbHandle->error) { $self->error($self->{dbHandle}->errorMessage); return undef; }
  if ($date eq $today)
  {
    use Finance::Quote;
    my $financeQuoteObj = Finance::Quote->new;
    my @tickers;
    #read array
    #walk array for tickers
    #use financequote to gather stock quotes
    $start = time;
    foreach my $h (@data)
    {
      if ($h->{exchange} ne '' && $h->{exchange} ne 'None')
      {
        push @tickers, $h->{ticker};
      }
    }
    if ($self->debug) { $self->debug('<div>Time to gather Data: '. (time - $start) .' seconds</div>'); }
    # fetch the information for the ticker symbols using Finance::Quote module
    # hardcode the use of the "usa" method for now
    # set the list of required labels that must exist for any stock quotes returned
    $financeQuoteObj->require_labels("name", "last", "date");
    $financeQuoteObj->failover(1);
    %quotes = $financeQuoteObj->fetch("usa", @tickers);
    # if nothing returned from the fetch, return an error message, but do not set an error condition
    if (defined %quotes)
    {
      my %index;
      my %accountTotals;
      my %totals = (mkt => '0.00', gain_loss => '0.00');
      #walk array and update price, market, gain_loss
      for (my $i = 0; $i < scalar @data; $i++)
      {
        my $h = $data[$i];
        if ($h->{level} <= 3)
        {
          $index{$i} = $h->{level};  #set marker on this row to flag that it needs to be updated
        }
        else
        {
          if ($date eq $today && $h->{exchange} ne '' && $h->{exchange} ne 'None')
          {
            $h->{price} = $quotes{$h->{ticker}, 'last'};
            $h->{mkt} = sprintf("%.2f", $h->{price} * $h->{shares});
            $h->{gain_loss} = sprintf("%.2f", $h->{mkt} - $h->{cost});
          }
          if ($h->{mkt} ne '')
          {
            $accountTotals{mkt}{$h->{account}} += $h->{mkt};
            $accountTotals{mkt}{$h->{account}.$h->{stock}} += $h->{mkt};
            $totals{mkt} += $h->{mkt};
          }
          if ($h->{gain_loss} ne '')
          {
            $accountTotals{gain_loss}{$h->{account}} += $h->{gain_loss};
            $accountTotals{gain_loss}{$h->{account}.$h->{stock}} += $h->{gain_loss};
            $totals{gain_loss} += $h->{gain_loss};
          }
  #         #format data for cleanup
  #         $h->{price} = sprintf('%.2f', $h->{price});
  #         $h->{mkt} = $self->commify(sprintf('%.2f',$h->{mkt}));
  #         $h->{gain_loss} = $self->commify(sprintf('%.2f',$h->{gain_loss}));
        }
      }
      for (my $i = 0; $i < scalar @data; $i++)
      {
        my $h = $data[$i];
        #check for account totals so that they can be updated
        if (defined $index{$i})  #look for marker on this row
        {
          if ($index{$i} == '1')
          {
            $h->{mkt} = $totals{mkt};
            $h->{gain_loss} = sprintf("%.2f",$totals{gain_loss});
          }
          elsif ($index{$i} == '2')
          {
            $h->{mkt} = $accountTotals{mkt}{$h->{account}};
            $h->{gain_loss} = sprintf("%.2f", $accountTotals{gain_loss}{$h->{account}});
          }
          elsif ($index{$i} == '3')
          {
            $h->{mkt} = $accountTotals{mkt}{$h->{account}.$h->{stock}};
            $h->{gain_loss} = sprintf("%.2f", $accountTotals{gain_loss}{$h->{account}.$h->{stock}} );
            $h->{cost_share} = sprintf('%.2f', $h->{cost}/$h->{shares});
            $h->{price} = $quotes{$h->{ticker}, 'last'};
          }
        }
        #convert whole row to html or text
        if ($format eq 'html'){ $table .= $self->HTMLconvert($h); }
        else { $table .= $self->TEXTconvert($h); }
      }
      if ($format eq 'html') { $r{body} .= '<table id="portfolio" class="portfolio">'."\n". $self->headers . $table . "</table>\n"; }
      else { $r{body} .= $self->textHeaders . $table; }
      return %r;
    }
    #if here then drop to default walkthrough data
    $r{body} .= '<div class="error">NO prices fetched from internet.'. "</div>\n";
  }
  #if here then the portfolio is not for today or prices failed.
  #walk data and transform
  for (my $i = 0; $i < scalar @data; $i++)
  {
    my $h = $data[$i];
    #convert whole row to html or text
    if ($format eq 'html'){ $table .= $self->HTMLconvert($h); }
    else { $table .= $self->TEXTconvert($h); }
  }

  if ($format eq 'html')
  {
    $r{body} .= '<table id="portfolio" class="portfolio">'."\n". $self->headers . $table . "</table>\n";
  }
  else
  {
    $r{body} .= $self->textHeaders . $table;
  }
  return %r;
}

sub headers
{
  my $self = shift;
  my $r = <<ENDOF;
    <thead>
ENDOF
  foreach my $field (@{$self->{columns}})
  {
    $r .= '    <th class="'.$field.' primary">'.$self->map($field)."</td>\n";
  }

  return $r."    </thead>\n";
}

sub textHeaders
{
  my $self = shift;
  my $r = "'level'";
  foreach my $field (@{$self->{columns}})
  {
    if ($r ne '') { $r .= ", "; }
    $r .= "'".$self->map($field)."'";
  }
  return $r."\n";
}

sub HTMLconvert
{
  my $self = shift;
  my $data = shift;
  my $r;
  foreach my $field (@{$self->{columns}})
  {
    my $d = $data->{$field};
    if ($field eq 'gain_loss')
    {
      my $f = $self->commify(sprintf("%.2f",$d));
      my $class = "zero";
      if ($d < 0.0) { $class="negative"; }
      elsif ($d > 0.0) { $class = 'positive'; }
      $d = "<span class=\"$class\">$f</span>\n";
    }
    elsif ($field eq 'cost' || $field eq 'mkt')
    {
      $d = $self->commify(sprintf("%.2f",$d));
    }
    elsif ($field eq 'shares')
    {
      $d = $self->commify($d);
    }
    elsif ($field eq 'price' && $d ne '')
    {
      $d = $self->commify(sprintf("%.2f",$d));
    }
    $r .= '    <td class="'.$field.'">'.$d."</td>\n";
  }
  if ($r ne '') { return '  <tr class="level'.$data->{level}.'">'."\n$r</tr>\n"; }
  return '';
}

sub TEXTconvert
{
  my $self = shift;
  my $data = shift;
  my $r = "'$data->{level}'";
  foreach my $field (@{$self->{columns}})
  {
    if ($r ne '') { $r .= ", "; }
    $r .= "'$data->{$field}'";
  }
  if ($r ne '') { $r .= "\n"; }
  return $r;
}

sub map
{
  my $self = shift;
  my $f = shift;
  return $self->langObj->map($self->{columnNames}->{$f});
}

1;
__END__

=back

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Base(3)

=cut
