# ReturnOfCapital.pm - The Object Class that provides a ReturnOfCapital Object
#$Id: ReturnOfCapital.pm,v 1.17.4.1 2006/08/08 17:12:46 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2005 http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Investments::Objects::ReturnOfCapital;
use strict;
use Portal::Accounting::Investments::Objects::Event;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Investments::Objects::Event Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.17.4.1 $';

=head1 NAME

ReturnOfCapital

=head1 SYNOPSIS

  use Portal::Accounting::Investments::Objects::ReturnOfCapital;
  my $obj = Portal::Accounting::Investments::Objects::ReturnOfCapital->new();
  $obj->populate(date =>, . . .  );
  if ($obj->error)
  {
    die $errorMessage();
  }

=head1 DESCRIPTION

  ReturnOfCapital is the class used to manage ReturnOfCapital security events.  It contains forms
  for creating, editing (if this gets implemented), and deleting ReturnOfCapital events.
  It also contains the database functionality for inserting and deleting
  records.  As with all security events LotNode is used as part of the
  processes to create, delete, and insert data.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(dbHandle => , langObj =>, ledgerCode =>)

  Creates a new instance of the Portal::Accounting::Investments::Objects::ReturnOfCapital module.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( ledgerCode => "", app => "Accounting", state => "Investments::Journal", @_ );

  # instantiate anything unique to this module
  $self->{dbHandle} = $args{dbHandle};
  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }
  # instantiate anything unique to this module
  my $me = "rtcp";
  my $l = $self->{langObj};
  my @date = $self->getDateArray;
  my $today = $self->getCurrentDate;

  $self->{defaultForm} = "main";
  $self->{data}->{main} = {
    "app" => { -Type => "hidden", -Value => $args{app} },
    "state" => { -Type => "hidden", -Value => $args{state} },
    "command" => { -Type => "hidden", -Value => "save" },
    "id" => { -Type => "hidden", },
    "counter" => { -Type => "hidden", },
    "counter" => { -Type => "hidden", -Value => "", -DataType => "int" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, -DataType => "string" },
    "type" => { -Type => "hidden", -Label => $l->map($me), -Value => $me, -DataType => "string" },
    "date" => { -Label => $l->map("Date"), -Type => "date-picker", -DataType => "string",
      -year => $date[0], -Value => $today },
    "stock" => { -Label => $l->map("Stock"), -Type => "select", -DataType => "int",
      -Options => $self->{stocks}      },
    "comment" => { -Type => "text", -DataType => "string", -Label => $l->map("Comment") },
    "total" => { -Type => "text", -DataType => "float", -Label => $l->map("reduceCost"), size => 8 },
  };
  $self->{data}->{report} = {
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, },
    "id" => { -Type => "hidden", },
    "counter" => { -Type => "hidden", },
    "type" => { -Type => "text", -Label => $l->map($me), -Value => $me, -DataType => "string", -ReadOnly => 1, },
    "date" => { -Label => $l->map("Date"), -Type => "text", -DataType => "string",
     -ReadOnly => 1, },
    "stock" => { -Label => $l->map("Stock"), -Type => "select", -DataType => "int",
      -Options => $self->{stocks}, -ReadOnly => 1,   },
    "comment" => { -Type => "text", -DataType => "string", -Label => $l->map("Comment"), -ReadOnly => 1, },
    "total" => { -Type => "text", -DataType => "float", -Label => $l->map("reduceCost"), -ReadOnly => 1, },
  };
  $self->{css}->{main} = <<"ENDOF";
ENDOF
  my $SelectLots = $l->map("SelectLots");
  my $Submit = $l->map("Submit");
  my $app = $self->{data}->{main}->{app}->{-Value};
  my $state = $self->{data}->{main}->{state}->{-Value};
  my $lc = $self->{data}->{main}->{ledgerCode}->{-Value};
  $self->{template}->{main} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIELD=app#
  #FIELD=state#
  #FIELD=command#
  #FIELD=counter#
  #FIELD=ledgerCode#
  <h4>#LABEL=type##FIELD=type#</h4>
  <table class="primary">
    <tr>
      <td>#LABEL=date# #FI=date#<br>
        #LABEL=stock# #FI=stock#
        #LABEL=total# \$<u>#FI=total#</u>.<br>
        #LABEL=comment# #FI=comment# <b><font color="green">%</font></b>
      </td>
    </tr>
  </table>
  <input type="submit" name="submit" value="$Submit" onclick="this.disabled = true;">
  <p><b><font color="green">%</font></b> Optional</p>
</form>
ENDOF
  $self->{template}->{report} = <<"ENDOF";
<table id="i#VALUE=id#" class="report">
  <tr>
    <td class="date">#FIELD=date#</td>
    <td class="event"><a title="#LABEL=type#">#FIELD=type#</a></td>
    <td class="detail">@ #FIELD=total# per share : #FIELD=comment# </td>
$self->{deleteString}   
  </tr>
</table>
ENDOF
  $self->{javascript}->{main} = <<"ENDOF";
ENDOF
  $self->{profile}->{main} = {
    required => [ qw( date total stock ) ],
    optional => [ qw( comment counter ) ],
    constraints => {
         date => qr/^\d{4}-\d{2}-\d{2}$/,
         total => 'RE_num_real',
         stock => qr/^\d+$/,
         counter => qr/^\d+$/,
       },
  };
  $self->{profile}->{report} = {
    required => [ qw( date total stock ) ],
    optional => [ qw( comment counter ) ],
    constraints => {
         date => qr/^\d{4}-\d{2}-\d{2}$/,
         total => 'RE_num_real',
         stock => qr/^\d+$/,
         counter => qr/^\d+$/,
       },
  };

  $self->{INSERTorder}[0] = [ qw( event_code trans_date counter s cc ledger_code event_ref comment ) ] ; #insert order for db column names
  
  $self->{db2data}[0] = {
    event_code => 'type',
    trans_date => 'date',
    s => 'stock',
    cc => 'total',
    ledger_code => 'ledgerCode',
    event_ref => 'id',
  };  #map database columns to ->data hash, not specified assumes same name
  $self->{data2db}[0] = {reverse %{$self->{db2data}[0]}};

  $self->{SELECT}[0] = "SELECT ##COLUMNS## FROM $self->{table} WHERE event_ref = #FIELD=id#";
#  $self->{INSERT}[0] = "INSERT INTO $self->{table} (##COLUMNS##) VALUES (##VALUES##)";
#  $self->{UPDATE}[0] = "UPDATE $self->{table} SET ##UPDATE## WHERE event_ref = #FIELD=id#";
  $self->{DELETE}[0] = "DELETE FROM $self->{table} WHERE event_ref = #FIELD=id#";

  
  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item boolean insert

  Inserts the record for this event into database and calls other functions to fill in other records. A buy might be affected by many events in the future.  PropagateDelete isn't used with buy because it isn't needed with a new node. There is no propagate routine in buy because it should never be called based on the way the system works.  A buy is only propagated during insert.
  
=cut
sub insert
{
  my $self = shift;
  if (!$self->isValid) { return undef; }
  my $eventRef = $self->nextSecurityEventRef();
  my $counter = $self->nextSecurityEventCounter($self->get('date'));
  $self->set('id', $eventRef);
  $self->set('counter', $counter);
  
  $self->{dbHandle}->write(
    sql => "INSERT INTO sec_event_tb (event_code, trans_date, counter, S, CC, ledger_code, event_ref, comment) "
      .                      "VALUES (?,          ?,          ?,       ?, ?,  ?,           ?,         ?)",
    plug => [
      $self->get('type'),
      $self->get('date'),
      $counter,
      $self->get('stock'),
      $self->get('total'),
      '*',
      $eventRef,
      $self->get('comment')
    ]
  );
  if (!$self->dbHandle->error)
  {
    # retrieve the affected lots for all ledgers
    $self->insertIntoLedgers(date => $self->get('date'), counter => $counter, stock => $self->get('stock'));
  }
  else
  {
    $self->error($self->dbHandle->errorMessage);
  }
  if (!$self->error) { $self->{dbHandle}->commit; }
  return $eventRef;
}

=item boolean propagate(LotNode)

  See Investments::Objects::Event::propagate

=cut
sub propagate
{
  my $self = shift;
  my $oldNode = shift;
  my $capitalGain = 0.0;
  my $returnAmount = sprintf("%.2f", $oldNode->get('shares') * $self->get('total') );
  my $returnCost = $returnAmount; #will be used for credit to stock sub account
  my $oldNodeNewCost = $oldNode->get('cost') - $returnAmount;

  # calculate the amount for the new cost (on a per share basis) and potentially capital gains
  if ($oldNodeNewCost < 0 ) # or ($returnAmount > $oldNode->get('cost') )
  {
    $capitalGain = sprintf("%.2f", abs($oldNodeNewCost) );
    $oldNodeNewCost = 0.00;
    $returnCost = $oldNode->get('cost');
  }

  # build a JournalEntry for the data
  my $transobj = Portal::Accounting::Objects::JournalEntry->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj}, ledgerCode => $oldNode->get('ledgerCode') );
  $transobj->set('journal', $self->{Journal}); # investements journal code hardcoded
  $transobj->set('date', $self->get('date') );
  $transobj->addRow(
    subEntity => $oldNode->get('subEntity'),
    account => $oldNode->get('account'),
    subAccount => $self->get('stock'),
    dorc => "false",
    amount => sprintf("%.2f", $returnCost),
    comment => $self->get('comment'),
  );
  $transobj->addRow(
    subEntity => $oldNode->get('subEntity'),
    account => $oldNode->get('account'),
    dorc => "true",
    amount => sprintf("%.2f", $returnAmount),
    comment => $self->get('comment'),
  );
  if ($capitalGain > 0.0)
  {
    # determine if the return of capital is occurring after one year that the security has been held in order to assign
    # long or short term gains or losses
    my $buydate;
    my $sql = "SELECT MIN(trans_date) from sec_event_tb WHERE l = (SELECT buy_lot FROM lot_node_tb WHERE node_id = ? LIMIT 1)";
    my $sth = $self->{dbHandle}->read(sql => $sql, plug => [$oldNode->get('id')] );
    if ($self->{dbHandle}->error())
    {
      $self->error($self->{dbHandle}->errorMessage);
      return 0;
    }
    my @info = $sth->fetchrow_array;
    $buydate = $info[0];
    my @ltst = $self->shortTermLongTerm($self->get('date'), $buydate);
    if ($self->error) { return 0; }

    $transobj->addRow(
      subEntity => $oldNode->get('subEntity'),
      account => $ltst[0],
      subAccount => $self->get('stock'),
      dorc => "false",
      amount => sprintf("%.2f", $capitalGain),
    );
  }
  if ($transobj->error)
  {
    $self->error($transobj->errorMessage);
    return 0;
  }
  # insert the transaction into the database
  my $refNum = $transobj->insert();
  if ($transobj->error)
  {
    $self->error($transobj->errorMessage);
    return 0;
  }

  # insert a new lot node into the database
  my $newLotNodeObj = Portal::Accounting::Investments::Objects::LotNode->new(ledgerCode => $oldNode->get('ledgerCode'), langObj => $self->langObj, dbHandle => $self->dbHandle);
  $newLotNodeObj->populate(
    date => $self->get('date'),
    parent => $oldNode->get('id'),
    event => $self->get('id'),
    journalRef => $transobj->get('id'),
    subEntity => $oldNode->get('subEntity'),
    account => $oldNode->get('account'),
    subAccount => $oldNode->get('subAccount'),
    lot => $oldNode->get('lot'),
    buyLot => $oldNode->get('buyLot'),
    cost => sprintf("%.2f", $oldNodeNewCost),
    shares => $oldNode->get('shares'),
  );
  if ($newLotNodeObj->error) { $self->error($newLotNodeObj->errorMessage); return 0; }
  $newLotNodeObj->insert;
  if ($newLotNodeObj->error) { $self->error($newLotNodeObj->errorMessage); return 0; }
  return $self->SUPER::propagate($newLotNodeObj);
}

=item boolean delete

  Routine calls all of the neccessary functions to get rid of this event and all nodes affected by it in lot trees.
  
=cut
sub delete
{
  my $self = shift;
  return $self->SUPER::deleteEvents(@_);
}

=back

=cut

1;
__END__

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Objects::Base(3)

=cut