#$Id: State.pm,v 1.1 2006/05/12 14:35:46 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.

package Portal::Accounting::Payroll::Objects::State;
use Portal::Accounting::Objects::BaseL;
use Portal::Accounting::Payroll::Objects::Employee;
use strict;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.1 $';

=head1 NAME

State - Object used to build a State Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Payroll::Objects::State;
  my $obj = Portal::Accounting::Payroll::Objects::State->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

State is a State class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::Payroll::Objects::State module.
 See Portal::Accounting::Objects::Base(3) for a listing of required arguments.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (ledgerCode => '', @_ );

  if ($self->error)
  {
    $self->error();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{name} = "Portal::Accounting::Payroll::Objects::State";
  $self->{new} = 1;  #after populate it is no longer new
  $self->{table} = "state_payroll";
  $self->{id} = 'id';
  my $l = $self->{langObj};

  $self->{periods} = {
    names => [qw(annual semiannual quarterly monthly semimonthly biweekly weekly daily)],
    values => [qw(annual semiannual quarterly monthly semimonthly biweekly weekly daily)],
  };
  $self->{states} = {
    MO => 'MO',
#    AZ => 'AZ',
    US => 'US',
  };
  $self->{EMPLOYEE} = Portal::Accounting::Payroll::Objects::Employee->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $args{ledgerCode});
  my $employees = $self->{EMPLOYEE}->selectOptions;
  my $employee;
  if (defined $args{employee}) { $employee = $args{employee}; }
  if ($self->{EMPLOYEE}->error) { $self->error($self->{EMPLOYEE}->errorMessage); }
  my $states = $self->{formObj}->createSelectOptions($self->{states} );
  # do anything else you might need to do.
  $self->{data}->{main} = {
    "app" => { -Type => "hidden", -Value => "Accounting" },
    "state" => { -Type => "hidden", -Value => "Payroll::State" },
    "command" => { -Type => "hidden", -Value => "save" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, -DataType => "string" },
    "states" => { -Type => 'select', -Options => $states, -DataType => "text", -Label => 'State', },
    "id" => {-Type => 'hidden', -DataType => 'int', -Value => 0 },
    "employee" => { -Type => 'select', -Options => $employees, -DataType => "integer", -Label => 'Employee', -Value => $employee},
    "key" => { -Type => 'text', -DataType => 'string', maxlength => 255, -Value => '',},
    "value" => { -Type => 'text', -DataType => 'string', maxlength => 255, -Value => '',},
  };
  $self->{css}->{main} .= <<"ENDOF";
ENDOF
  $self->{template}->{main} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIR=app#
  #FIR=state#
  #FIR=command#
  #FIR=ledgerCode#
  <div class="primary">
    <div>#FORMREQUIREDSTRING#</div>
    <div>#LABEL=id# #FIR=id#</div>
    <div>#LABEL=states# #FIR=states#</div>
    <div>#LABEL=employee# #FIR=employee#</div>
    <div>#LABEL=key# #FIR=key#</div>
    <div>#LABEL=value# #FIR=value#</div>
  </div>
</form>  
ENDOF
  $self->{javascript}->{main} = <<"ENDOF";
ENDOF
  $self->{profile}->{main} = {
    required => [ qw( ) ],
    optional => [ qw( ) ],
    constraints => {
    },
  };
  $self->{profile}->{list} = {
    required => [ qw( ) ],
    optional => [ qw( ) ],
    constraints => {
    },
  };
  $self->{profile}->{selector} = {
    required => [ qw(states employee ) ],
    optional => [ qw( ) ],
    constraints => {
    },
  };
  
  $self->{data}->{list} = {
    "app" => { -Type => "hidden", -Value => "Accounting" },
    "state" => { -Type => "hidden", -Value => "Payroll::State" },
    "command" => { -Type => "hidden", -Value => "delete" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, -DataType => "string" },
    "states" => { -Type => 'hidden', -DataType => "text",},
    "id" => {-Type => 'hidden', -DataType => 'int', -Value => 0 },
    "employee" => { -Type => 'hidden', -DataType => "integer", },
    "key" => { -Type => 'hidden', -DataType => 'string', -Value => '',},
    "value" => { -Type => 'hidden', -DataType => 'string', -Value => '',},
    "submitButton" => {-Type => 'button', -Value => "X", title => 'Delete', onclick => 'deleteStatePayroll(this.form.id.value);'},
  };
  $self->{headers}->{list} = <<"ENDOF";
<tr id="p#VALUE=id#">
  <th>State</th>
  <th>Key</th>
  <th>Value</th>
  <th>Employee</th>
  <th>Delete</th>
</tr>
ENDOF
  $self->{template}->{list} = <<"ENDOF";
<tr id="p#VALUE=id#">
  <td>#VALUE=states#</td>
  <td>#VALUE=key#</td>
  <td>#VALUE=value#</td>
  <td>#VALUE=employee#</td>
  <td>  
  <form #FORMARGS#>
    #FIR=app#
    #FIR=state#
    #FIR=command#
    #FIR=ledgerCode#
    #FIELD=id#
    #FIELD=states#
    #FIELD=employee#
    #FIELD=key#
    #FIELD=value#
    #FIELD=submitButton#
  </form>
  </td>
</tr>
ENDOF
  
  $self->{data}->{selector} = {
    "app" => { -Type => "hidden", -Value => "Accounting" },
    "state" => { -Type => "hidden", -Value => "Payroll::State" },
    "command" => { -Type => "hidden", -Value => "new" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}, -DataType => "string" },
    "states" => { -Type => 'select', -Options => $states, -DataType => "text", -Label => 'State', },
    "employee" => { -Type => 'select', -Options => $employees, -DataType => "integer", -Label => 'Employee', -Value => $employee},
  };
  $self->{css}->{selector} .= <<"ENDOF";
ENDOF
  $self->{template}->{selector} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIR=app#
  #FIR=state#
  #FIR=command#
  #FIR=ledgerCode#
  <div class="primary">
    Please choose an employee and the payroll module.  This module will help you setup the proper configuration.
    <div>#LABEL=states# #FIR=states#</div>
    <div>#LABEL=employee# #FIR=employee#</div>
    <input type="submit" value="Next">
  </div>
</form>  
ENDOF
  $self->{javascript}->{selector} = <<"ENDOF";
ENDOF

  $self->{LIST}[0] = "SELECT key||' = '||value, id FROM $self->{table} WHERE employee IN (SELECT id FROM employees_tb WHERE ledger_code = '$args{ledgerCode}') ORDER BY state, key, employee";
  if (exists $args{employee})
  {
    $self->{LIST}[0] = "SELECT key||' = '||value, id FROM $self->{table} WHERE employee = '#FIELD=employee#' ORDER BY state, key";
  }
  $self->{INSERTorder}[0] = [ qw( id state employee key value ) ] ;  #insert order for db column names
  $self->{newOnly}[0] = { qw(id) };
  $self->{db2data}[0] = { state => 'states' };  #map database columns to ->data hash, not specified assumes same name
  $self->{data2db}[0] = {reverse %{$self->{db2data}[0]}};
  $self->{SELECT}[0] = "SELECT ##COLUMNS## FROM $self->{table} WHERE id = '#FIELD=id#'";
  $self->{INSERT}[0] = "INSERT INTO $self->{table} (##COLUMNS##) VALUES (##VALUES##)";
  $self->{UPDATE}[0] = "UPDATE $self->{table} SET ##UPDATE## WHERE $self->{id} = '#FIELD=$self->{id}#'";

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  if ($self->get('ledgerCode') eq '') { $self->missing('ledgerCode'); }
  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub isNew { my $self = shift; return $self->{new}; }

sub printHTMLHeader
{
  my $self = shift;
  return $self->{headers}->{list};
}

sub printHTMLRow
{
  my $self = shift;
  my %form = $self->printForm(form => 'list');
  return $form{body};
}

sub insert
{
  my $self = shift;
  my $sql = "DELETE FROM $self->{table} WHERE state = ? AND employee = ?";
  my @plug = ($self->get('states'), $self->get('employee'));
  my $sth = $self->dbHandle->write(sql => $sql, plug => \@plug );
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return undef; }
  return $self->SUPER::insert(@_);
}

=back

=cut

1;
__END__

=head1 NOTE

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::Objects::Base(3)

=cut
