package Portal::Accounting::Reports::Display;
use strict;
use Portal::Accounting::Reports::Main;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);
use JSON;
use HTMLObject::Base;
use HTMLObject::ReadCookie; # the module used to read cookies sent to us from the client.
use Storable;  # the module that will allow me to dump data to disk and recover later.  The file names will be stored in the cookie.
use POSIX qw(tmpnam); # the tmpnam function which will return a file name for a temporary file supposed to be unique.
# Program Modes:  Uses command
# command =
#   = 'run_report_top' - Generates the interface to allow the user to customize and print the report and memorize the current settings.
#   = 'run_report_bottom' - Actually displays the report based upon the current settings and user actions.

require Exporter;

@ISA = qw(Portal::Accounting::Reports::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.25 $';

my %commands = (
  display => "Reports",       #does nothing
  run_report_top => "Reports",
  run_report_bottom => "Reports",
);

my %errors;
%errors = ( 4000 => 'Eval of HTMLObject::FrameSet failed',
        4001 => 'Invalid command specified',
        4002 => 'Eval of HTMLObject::Normal failed',
        4003 => 'command must be specified',
        4004 => "Couldn't connect to Database",
        4005 => "Parameter requires a value",
        4006 => "Invalid Value for Parameter",
        4007 => "DBI Preparation of SQL Statement Failed",
        4008 => "DBI Execution of Statement Failed",
        4009 => "DBI Error Occurred",
        4018 => "Error opening file",
        4028 => "Database Handle required",
        4035 => "File Name required",
        4036 => "Report Type required",
      );

my @font_sizes = ( "+2", "+1", "+0", "-1", "-2" );
my %font_sizes = ( "+2" => "+2", "+1" => "+1", "+0" => "nothing", "-1" => "-1", "-2" => "-2" );
my %report_classes;
%report_classes = ( "custom" => "Custom", "cash" => "Cash Flow", "income" => "Income Statement", "balance" => "Balance Sheet", "acct_balance" => "Accounts Balance Sheet", "acct_activity" => "Accounts Activity Sheet" );
my @report_classes;
@report_classes = ( "custom", "cash", "income", "balance", "acct_balance", "acct_activity" );
my @account_type_levels;
@account_type_levels = ( "nothing", "accounts", "accounts, sub-accounts" );
my @account_levels;
@account_levels = ( "nothing", "sub-accounts" );
my @accounts_of_levels;
@accounts_of_levels = ( "nothing", "sub-accounts" );
my @sub_account_levels;
@sub_account_levels = ( "nothing", "accounts" );
my @sub_accounts_of_type_levels;
@sub_accounts_of_type_levels = ( "nothing", "accounts" );
my @sub_accounts_of_account_levels;
@sub_accounts_of_account_levels = ( "not available" );
my %regular_cell_types;
%regular_cell_types = ( "text" => "Text", "total" => "Total", "compare" => "Compare Rows", "account_type" => "Account Type", "account" => "Account", "sub_account" => "Sub Account", "accounts_of" => "Accounts of Account Type", "sub_accounts_of_type" => "Sub Accounts of Account Type", "sub_accounts_of_account" => "Sub Accounts of Account", "begin_account_type_balance" => "Beginning Account Type Balance", "end_account_type_balance" => "Ending Account Type Balance", "begin_account_balance" => "Beginning Account Balance", "end_account_balance" => "Ending Account Balance", "begin_sub_account_balance" => "Beginning Sub Account Balance", "end_sub_account_balance" => "Ending Sub Account Balance", "line" => "Line" );
my @regular_cell_types;
@regular_cell_types = ( "text", "total", "compare", "account_type", "account", "sub_account", "accounts_of", "sub_accounts_of_type", "sub_accounts_of_account", "begin_account_type_balance", "end_account_type_balance", "begin_account_balance", "end_account_balance", "line" ); # "begin_sub_account_balance", "end_sub_account_balance", (to be added later)
my %column_names;
%column_names = ( 0 => "Zero", 1 => "One", 2 => "Two", 3 => "Three", 4 => "Four", 5 => "Five", 6 => "Six", 7 => "Seven", 8 => "Eight", 9 => "Nine", 10 => "Ten", 11 => "Eleven", 12 => "Twelve", 13 => "Thirteen", 14 => "Fourteen", 15 => "Fifteen", 16 => "Sixteen", 17 => "Seventeen", 18 => "Eighteen", 19 => "Nineteen", 20 => "Twenty", 21 => "Twenty One", 22 => "Twenty Two", 23 => "Twenty Three", 24 => "Twenty Four", 25 => "Twenty Five", 26 => "Twenty Six", 27 => "Twenty Seven", 28 => "Twenty Eight", 29 => "Twenty Nine", 30 => "Thirty" );
my @column_headers;
@column_headers = ( "sub_entities", "sub_accounts", "sub_account_types" );
my %column_headers;
%column_headers = ( "sub_entities" => "Sub Entities", "sub_accounts" => "Sub Accounts", "sub_account_types" => "Sub Account Types" );
my @negative_formats;
@negative_formats = ( "0", "1", "2" );
my %negative_formats;
%negative_formats = ( "0" => "<font color=\"red\">(xxx.xx)</font>", "1" => "<font color=\"red\">-</font>xxx.xx", "2" => "xxx.xx<font color=\"red\">-</font>" );
my $max_dates;
$max_dates = 4;   # the maximum number of columns that we are currently supporting in the reporting system.
my %t_or_f;
%t_or_f = ( 0 => "false", 1 => "true", "true" => "true", "false" => "false" );  # a mapping that will allow us to map 0 to false and 1 to true for the sql statements.
my @precision_formats;
@precision_formats = ( "1", "10", "100", "1000", "10000", "100000", "1000000", "1000000000" );
my %precision_formats;
%precision_formats = ( "1" => "1", "10" => "10", "100" => "100", "1000" => "1,000", "10000" => "10,000", "100000" => "100,000", "1000000" => "1,000,000", "1000000000" => "1,000,000,000" );
my %precision_names;
%precision_names = ( "10" => "tens", "100" => "hundreds", "1000" => "thousands", "10000" => "ten-thousands", "100000" => "hundred-thousands", "1000000" => "millions", "1000000000" => "billions" );
my %true_false;
%true_false = ( "true" => 1, "false" => 0, 1 => 1, 0 => 0 );

# commonly used compiled regular expressions
my $tripleAccountsRE = qr/^(account_type|account|accounts_of)$/;
my $tripleBeginAccountsRE = qr/^(begin_account_type_balance|begin_account_balance|begin_sub_account_balance)$/;
my $tripleEndAccountsRE = qr/^(end_account_type_balance|end_account_balance|end_sub_account_balance)$/;
my $singleQuoteRE = qr/\'/;
my $negativeOneRE = qr/^(-1)$/;
my $dotRE = qr/\./;

# regular view display settings values
my %hide_numbers;
%hide_numbers = {};
$hide_numbers{'account_type'} = {};
$hide_numbers{'account_type'}{'0'} = 1;

$hide_numbers{'accounts_of'} = {};
$hide_numbers{'accounts_of'}{'0'} = 1;

$hide_numbers{'begin_account_type_balance'} = {};
$hide_numbers{'begin_account_type_balance'}{'0'} = 1;

$hide_numbers{'end_account_type_balance'} = {};
$hide_numbers{'end_account_type_balance'}{'0'} = 1;

$hide_numbers{'sub_accounts_of_type'} = {};
$hide_numbers{'sub_accounts_of_type'}{'0'} = 1;
$hide_numbers{'sub_accounts_of_type'}{'1'} = 1;

# my $cookies;
# $cookies = HTMLObject::ReadCookie->new;
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (@_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  if (defined $self->{input}->{ledgerCode}) { $self->{ledgerCode} = $self->{input}->{ledgerCode}; }
  else { $self->error('no ledgerCode.  Reports do not yet support running consolidated Ledgers');  }
  
  # do validation
  if (!$self->Portal::Accounting::Reports::Display::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  $self->{commands} = \%commands;

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }
  
  # validate our parameters.
  if (!defined $self->{input}->{file_name})
  {
    $self->missing();
  }
  if (!defined $self->{input}->{report_type})
  {
    $self->missing();
  }
  if ($self->numInvalid() > 0 ,, $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }
  return 1;
}

sub prepDoc
{
  my $self = shift;
  my $message;
  if (defined $self->{input}->{MESSAGE}) { $message = $self->{input}->{MESSAGE}; }
  my %args = (doc => HTMLObject::Normal->new(), @_);
  my $color = $self->{sessionObj}->{store}->{ledgers}->{$self->{input}->{ledgerCode}}->{color};
  my %foo;
  my $doc;
  if (defined $args{doc}) { $doc = $args{doc}; }
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->error)
  {
    $self->prefixError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->error("No valid colorScheme returned!<br>\n");
    return undef;
  }
  my %cssEntries = $self->{methods}->generateColorSchemeCSS(colorScheme => $colorScheme, appName => $self->{app});
  if ($self->{methods}->error)
  {
    $self->error($self->{methods}->errorMessage);
    return undef;
  }
  $doc->print(%cssEntries);
  $doc->print(css => [ ".ledgerFGColor { color: $color; }", ".ledgerBGColor { background-color: $color; }", ".ledgerBorderColor { border: none; border-style: solid ; border-color: $color; }" ]);
  $doc->print(css => [ ".title { text-align: center; font-weight: bold; }" ]);
  $doc->print(css => [ ".text { text-align: left; }" ]);
  $doc->print(css => [ ".numbers { text-align: right; }" ]);
  $doc->setStyleEntry(tag => 'a', string => 'color: #000000; ');
  $doc->setStyleEntry(tag => 'a:link', string => 'color: #000000; ');
  $doc->setStyleEntry(tag => 'a:active', string => 'color: #000000; ');
  $doc->setStyleEntry(tag => 'a:hover', string => 'color: #000000; ');
  $doc->setJavascriptInclude(file => $self->{jsDir}."Accounting.js");
  $doc->setJavascriptInclude(file => $self->{jsDir}."Reports/Display.js");
  return $doc;
}

sub c_run_report_top
{
  my $self = shift;
  my $doc;
  my $file_name = $self->{input}->{file_name};
  my $report_type = $self->{input}->{report_type};
  # gather the stored data from disk.
  if ($report_type eq "r")
  {
    $doc = $self->display_regular_report_top(file_name => $file_name );
  }
  return $doc;
}

sub c_run_report_bottom
{
  my $self = shift;
  my %input = %{$self->{input}};
  my $doc = $self->prepDoc;
  my $file_name = $self->{input}->{file_name};
  my $report_type = $self->{input}->{report_type};

  if ($report_type ne "r")
  {
    $doc->print('<div class="error">Invalid report type: '.$report_type."</div>\n");
    return $doc;
  }
  my @store_data = ();
  my $debugStr = "";
  my $args_ref;
  $args_ref = retrieve($file_name);
  my %args;
  my @data = ();
  %args = %{@{$args_ref}[0]};
  @data = @{@{$args_ref}[1]};
  my $elapsedTime = @{$args_ref}[2];
  my $id = $args{'id'};
  my $ledgerCode = $self->{input}->{'ledgerCode'};

  # set to store original settings before any modifications to them take place
  $store_data[0] = { %args };
  $store_data[2] = $elapsedTime;

  # update our displayZeroRows flag.
  $args{displayZeroRows} = (exists $input{displayZeroRows} ? $input{displayZeroRows} : $args{displayZeroRows});

  if (exists $input{'report_command'} && $input{'report_command'} !~ /^(print|export)$/)
  {
    my $level = $input{'level'};
    my $row = $input{'row'};
    (my $sub_row1, my $sub_row2, my $sub_row3) = split /\|/, $input{'sub_row'}, 3;

    if ($input{'report_command'} eq "open")
    {
      if ($sub_row1 == -1)
      {
        if (defined $data[$row])
        {
          $self->expandRow($data[$row], $args{displayZeroRows}, -3, [], \$debugStr);
        }
      }
      elsif ($sub_row1 ne "" && $sub_row2 eq "" && $sub_row3 eq "")
      {
        if (defined $data[$row]{'sub_rows'}{$sub_row1})
        {
          $self->expandRow($data[$row], $args{displayZeroRows}, -3, [ $sub_row1 ], \$debugStr);
        }
      }
      elsif ($sub_row1 ne "" && $sub_row2 ne "" && $sub_row3 eq "")
      {
        if (defined $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows'}{$sub_row2})
        {
          $self->expandRow($data[$row], $args{displayZeroRows}, -3, [ $sub_row1, $sub_row2 ], \$debugStr);
        }
      }
      elsif ($sub_row1 ne "" && $sub_row2 ne "" && $sub_row3 ne "")
      {
        if (defined $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows'}{$sub_row2}{'sub_rows'}{$sub_row3})
        {
          $self->expandRow($data[$row], $args{displayZeroRows}, -3, [ $sub_row1, $sub_row2, $sub_row3 ], \$debugStr);
        }
      }
    }
    elsif ($input{'report_command'} eq "close")
    {
      if ($sub_row1 == -1)
      {
        if (defined $data[$row])
        {
          $data[$row]{'sub_rows_visible'} = "false";
        }
      }
      elsif ($sub_row1 ne "" && $sub_row2 eq "" && $sub_row3 eq "")
      {
        if (defined $data[$row]{'sub_rows'}{$sub_row1})
        {
          $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows_visible'} = "false";
        }
      }
      elsif ($sub_row1 ne "" && $sub_row2 ne "" && $sub_row3 eq "")
      {
        if (defined $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows'}{$sub_row2})
        {
          $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows'}{$sub_row2}{'sub_rows_visible'} = "false";
        }
      }
      elsif ($sub_row1 ne "" && $sub_row2 ne "" && $sub_row3 ne "")
      {
        if (defined $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows'}{$sub_row2}{'sub_rows'}{$sub_row3})
        {
          $data[$row]{'sub_rows'}{$sub_row1}{'sub_rows'}{$sub_row2}{'sub_rows'}{$sub_row3}{'sub_rows_visible'} = "false";
        }
      }
    }
    if ($input{'report_command'} eq "reset")
    {
      $level = "-4";
      for (my $rowCount=0; $rowCount < scalar @data; $rowCount++)
      {
        my $row = $data[$rowCount];
        # always show the base level for the row; a defined row of the report will always be shown no matter what its level
        # unless it has been defined as an invisible row
        if ($row->{'show'} eq "1")
        {
          $row->{'visible'} = "true";
        }
        else
        {
          $row->{'visible'} = "false";
        }
        if (defined $row->{'sub_rows'})
        {
          foreach my $subCount1 (keys %{$row->{'sub_rows'}})
          {
            my $subRow1 = $row->{sub_rows}{$subCount1};
            if ($subRow1->{'initial_state'} eq "o")
            {
              $row->{'sub_rows_visible'} = "true";
              $subRow1->{'visible'} = "true";
            }
            else
            {
              $row->{'sub_rows_visible'} = "false";
              $subRow1->{'visible'} = "false";
            }
            if (defined $subRow1->{'sub_rows'})
            {
              foreach my $subCount2 (keys %{$subRow1->{'sub_rows'}})
              {
                my $subRow2 = $subRow1->{sub_rows}{$subCount2};
                if ($subRow2->{'initial_state'} eq "o")
                {
                  $subRow1->{'sub_rows_visible'} = "true";
                  $subRow2->{'visible'} = "true";
                }
                else
                {
                  $subRow1->{'sub_rows_visible'} = "false";
                  $subRow2->{'visible'} = "false";
                }
                if (defined $subRow2->{'sub_rows'})
                {
                  foreach my $subCount3 (keys %{$subRow2->{'sub_rows'}})
                  {
                    my $subRow3 = $subRow2->{sub_rows}{$subCount3};
                    if ($subRow3->{'initial_state'} eq "o")
                    {
                      $subRow2->{'sub_rows_visible'} = "true";
                      $subRow3->{'visible'} = "true";
                    }
                    else
                    {
                      $subRow2->{'sub_rows_visible'} = "false";
                      $subRow3->{'visible'} = "false";
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    $level = "-4" if ($input{report_command} eq "display");
    if ($input{'report_command'} =~ /^(expand|display|reset)$/)
    {
      # if we are displaying and the user has specified !displayZeroRows then we need to toggle visibility.
      for (my $rowCount=0; $rowCount < scalar @data; $rowCount++)
      {
        my $row = $data[$rowCount];
        $self->expandRow($row, $args{displayZeroRows}, $level, [], \$debugStr);
      }
    }
  }

  # store the (potentially) re-arranged data back to the same file; store before modifying original arguments but after
  # modifying opened and closed levels
  $store_data[1] = [ @data ];
  store(\@store_data, $file_name);

  my $max_display_level = $args{'max_display_level'};

  $doc->print("Debug Info:<br>\n${debugStr}") if (${debugStr});

  my @arguments = ( "report_name", "account_format", "negative_format", "data_font", "data_size", "text_font", "text_size", "total_column", "underline_length", "border", "pageTop", "reportWidth" );

  foreach my $value (@arguments)
  {
    if (exists $input{$value})
    {
      if (length $input{$value} == 0 && ($value ne "headers" && $value ne "sorted_by" && $value ne "sorted_order"))
      {
        $self->error("Error: Report.cgi - display_detailed_report. $errors{4005} Parameter = '$value'");
        return undef;
      }
      if ($value eq "pageTop" && $input{$value} < 0)
      {
        $input{$value} = 0;
      }
      $args{$value} = $input{$value};  # override the arguments passed in with the values from the web page.
    }
    else
    {
      if ($value eq "pageTop")
      {
        $args{$value} = 0;
      }
    }
  }

  my $with_links = "true";
  if ($input{'report_command'} =~ /^(print|export)$/)
  {
    $with_links = "false";
  }

  my $withUnderlines = 1;
  if ($input{report_command} eq "export")
  {
    $withUnderlines = 0;
  }

  # display the ledger name and the current date
  my $current_date = POSIX::strftime('%Y-%m-%d', localtime);
  chomp $current_date;

  $doc->setTitle("$args{'ledger_name'} - $args{'report_name'} - run on $current_date");

  $doc->setFocus("body");
#  $doc->setLinkDecorations();

  my $tableWidth = $args{reportWidth};

  if ($with_links eq "true")  # only output the time if not printing
  {
    $doc->print("<font size=\"-1\">Report took <b>$elapsedTime</b> seconds to gather data...</font><br>\n");
    $tableWidth = "100";
  }

  my $zero = '  <input type="hidden" name="displayZeroRows" value="'.$args{displayZeroRows}.'">';
  $doc->print(<<"END_OF_BODY");
<div style="text-align: center">
  <form method="post" name="run_form" onSubmit="document.run_form.pageTop.value=window.pageYOffset; return true;">
  $self->{postInfo}
  <input type=hidden name="command" value="run_report_bottom">
  <input type=hidden name="report_type" value="$report_type">
  <input type=hidden name="file_name" value="$file_name">
  <input type=hidden name="report_command" value="">
  <input type=hidden name="report_name" value="$args{'report_name'}">
  <input type=hidden name="account_format" value="$args{'account_format'}">
  <input type=hidden name="negative_format" value="$args{'negative_format'}">
  <input type=hidden name="text_font" value="$args{'text_font'}">
  <input type=hidden name="text_size" value="$args{'text_size'}">
  <input type=hidden name="data_font" value="$args{'data_font'}">
  <input type=hidden name="data_size" value="$args{'data_size'}">
  <input type=hidden name="total_column" value="$args{'total_column'}">
  <input type=hidden name="underline_length" value="$args{'underline_length'}">
  <input type=hidden name="border" value="$args{'border'}">
  <input type=hidden name="level" value="">
  <input type=hidden name="row" value="">
  <input type=hidden name="sub_row" value="">
  <input type=hidden name="pageTop" value="0">
  <input type="hidden" name="reportWidth" value="$args{reportWidth}">
  $zero
  <table border="$args{'border'}" cellspacing=0 cellpadding=1 width="$tableWidth%">
    <thead>
    <tr>
END_OF_BODY
  my @column_data = split /,/, $args{'column_data'};
  my @column_data_names = split /\|/, $args{'column_data_names'};
  my %name_num;
  foreach my $name_value (@column_data_names)
  {
    (my $column_name, my $column_num) = split " - ", $name_value;
    $name_num{$column_num} = $column_name;
  }

  for (my $count=0; $count < scalar @column_data; $count++)
  {
    if ($column_data[$count] eq "<non_sub_entity>")
    {
      $column_data[$count] = "";
    }
  }

  my $colSpan = (($args{'total_column'} eq "true" ? 1 : 0) + (($max_display_level == -1 && $args{'column_type'} ne "nothing") ? 1 : $max_display_level + 1) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1))));
  $doc->print(<<"END_OF_BODY");
      <td colspan="$colSpan" align="center" valign="top">
        <span style="font-weight: bold;"><font size="$args{'text_size'}">$args{'report_name'}</font></span>
      </td>
    </tr>
END_OF_BODY
  if ($args{'data_precision'} > 1)
  {
    $doc->print("    <tr>\n");
    $doc->print("      <td align=right valign=top colspan=" . (($args{'total_column'} eq "true" ? 1 : 0) + (($max_display_level == -1 && $args{'column_type'} ne "nothing") ? 1 : $max_display_level + 1) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1)))) . "><font size=\"$args{'data_size'}\" face=\"$args{'data_font'}\">(In <i>$precision_names{$args{'data_precision'}}</i>)</font></td>\n");
    $doc->print("    </tr>\n");
  }
  $doc->print(<<"END_OF_BODY");
    <tr>
END_OF_BODY
  if (($max_display_level == -1 && $args{'column_type'} ne "nothing") || $max_display_level > -1)
  {
    $doc->print('      <td colspan="' . (($max_display_level + 1 == 0 && $args{'column_type'} eq "nothing") ? 1 : $max_display_level + 1) . "\"><br></td>\n");
  }
  for (my $header_count=0; $header_count < scalar @{$args{'dates_entered'}}; $header_count++)
  {
    (my $date1, my $date2) = split /,/, $args{'dates_entered'}[$header_count];
    my $label = $args{"column$header_count"."_label"};
    $doc->print("      <td align=right valign=top");
    if (scalar @column_data > 0)
    {
      $doc->print(" colspan=" . (scalar @column_data));
    }
    $doc->print("><font size=\"$args{'text_size'}\" face=\"$args{'text_font'}\">$label</font></td>\n");
  }
  if ($args{'total_column'} eq "true")
  {
    $doc->print("      <td align=center valign=top><font size=\"$args{'text_size'}\" face=\"$args{'text_font'}\"><b>Total</b></font></td>\n");
  }
  $doc->print(<<"END_OF_BODY");
    </tr>
    </thead>
END_OF_BODY
  if ($args{'column_type'} ne "nothing")
  {
    $doc->print("    <tr>\n");
    $doc->print("      <td align=right valign=top");
    if ($max_display_level > -1)
    {
      $doc->print(" colspan=" . ($max_display_level + 1));
    }
    $doc->print(" nowrap><font size=\"$args{'data_size'}\" face=\"$args{'data_font'}\">$column_headers{$args{'column_type'}}:</font></td>\n");
    for (my $header_count=0; $header_count < scalar @{$args{'dates_entered'}}; $header_count++)
    {
      for (my $column_count=0; $column_count < scalar @column_data; $column_count++)
      {
        $name_num{$column_data[$column_count]} =~ s/\</\&lt\;/g;
        $name_num{$column_data[$column_count]} =~ s/\>/\&gt\;/g;
        $doc->print("      <td align=center valign=top><font size=\"$args{'data_size'}\" face=\"$args{'data_font'}\">");
        my $hide_number = 0;
        if ($args{'column_type'} eq "sub_account_types")
        {
          $hide_number = 1;
        }
        $doc->print($self->format_account(hide_number => $hide_number, account_format => $args{'account_format'}, name => $name_num{$column_data[$column_count]}, code => $column_data[$column_count]));
        $doc->print("</font></td>\n");
      }
    }
    if ($args{'total_column'} eq "true")
    {
      $doc->print("      <td><br></td>\n");
    }
    $doc->print("    </tr>\n");
  }

  my $error_occurred = 0;

  for (my $row_count=0; $row_count < scalar @data; $row_count++)
  {
    my $row = $data[$row_count];
    my $text_font = $args{'text_font'};
    my $text_size = $args{'text_size'};
    my $data_font = $args{'data_font'};
    my $data_size = $args{'data_size'};
    my $cell_type = $row->{'cell_type'};
    my $negative_format;
    my $underline_length = $args{'underline_length'};
    my $account_format = $args{'account_format'};
    my $rowVisible = $row->{visible};
    my $rowBold = $row->{bold};
    my $rowDrillable = $row->{drillable};
    my $rowTotal = $row->{total};

    if ($args{'negative_format'} == -1)
    {
      $negative_format = $row->{'negative_format'};
    }
    else
    {
      $negative_format = $args{'negative_format'};
    }

    if ($cell_type eq "text")
    {
      if ($rowVisible eq "true")
      {
        $doc->print("    <tr>\n");
        $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\"><font size=\"$text_size\" face=\"$text_font\">");
        for (my $skip_count=0; $skip_count < $row->{'display_level'}; $skip_count++)
        {
          $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
        }
        if ($rowBold eq "1")
        {
          $doc->print("<b>");
        }
        if ($row->{'cell_text'} ne "")
        {
          $doc->print("$row->{'cell_text'}");
        }
        else
        {
          $doc->print("&nbsp;");
        }
        if ($rowBold eq "1")
        {
          $doc->print("</b>");
        }
        $doc->print("</font></td>\n");
        $doc->print('      <td class="numbers" colspan=' . (($args{'total_column'} eq "true" ? 1 : 0) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1)))) . "><br></td>\n");
        $doc->print("    </tr>\n");
      }
    }
    elsif ($cell_type eq "line")
    {
      if ($rowVisible eq "true")
      {
        $doc->print("    <tr>\n");
        $doc->print("      <td colspan=" . (($max_display_level + 1) + ($args{'total_column'} eq "true" ? 1 : 0) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1)))) . " nowrap><font size=\"$text_size\" face=\"$text_font\">");
        for (my $skip_count=0; $skip_count < $row->{'display_level'}; $skip_count++)
        {
          $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
        }
        $doc->print("<hr noshade width=\"100%\"></font></td>\n");
        $doc->print("    </tr>\n");
      }
    }
    elsif ($cell_type eq "total" || $cell_type eq "compare")
    {
      if ($rowVisible eq "true")
      {
        $doc->print("    <tr>\n");
        $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\"><font size=\"$text_size\" face=\"$text_font\">");
        for (my $skip_count=0; $skip_count < $row->{'display_level'}; $skip_count++)
        {
          $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
        }
        if ($rowBold eq "1")
        {
          $doc->print("<b>");
        }
        if ($row->{'cell_text'} ne "")
        {
          $doc->print("$row->{'cell_text'}");
        }
        else
        {
          $doc->print("&nbsp;");
        }
        if ($rowBold eq "1")
        {
          $doc->print("</b>");
        }
        $doc->print("</font></td>\n");
        for (my $value_count=0; $value_count < scalar @{$row->{'values'}}; $value_count++)
        {
          my $valueCol = $row->{values}[$value_count];
          for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
          {
            $doc->print('      <td class="numbers" style="padding-right: ' . ($row->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
            if ($rowBold eq "1")
            {
              $doc->print("<b>");
            }
            if ($valueCol->[$value_count2] < 0)
            {
              if ($row->{underline} == 1 && $withUnderlines)
              {
                $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
              }
              else
              {
                $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
              }
            }
            else
            {
              if ($cell_type eq "total")
              {
                if ($row->{underline} == 1 && $withUnderlines)
                {
                  $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                }
                else
                {
                  $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                }
              }
              else
              {
                if ($row->{result_type} eq "result_formula")
                {
                  if ($row->{underline} == 1 && $withUnderlines)
                  {
                    $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                  }
                  else
                  {
                    $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                  }
                }
                else
                {
                  $doc->print($valueCol->[$value_count2]);
                }
              }
            }
            if ($rowBold eq "1")
            {
              $doc->print("</b>");
            }
            #$doc->print("<br>\n" . $row->{'debug_formula'}[$value_count][$value_count2]);
            $doc->print("</font></td>\n");
          }
        }
        if ($args{'total_column'} eq "true")
        {
          $doc->print('      <td class="numbers" style="padding-right: ' . ($row->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
          if ($rowBold eq "1")
          {
            $doc->print("<b>");
          }
          if ($rowTotal < 0)
          {
            if ($row->{underline} == 1 && $withUnderlines)
            {
              $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
            }
            else
            {
              $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
            }
          }
          else
          {
            if ($cell_type eq "total")
            {
              if ($row->{underline} == 1 && $withUnderlines)
              {
                $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
              }
              else
              {
                $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
              }
            }
            else
            {
              if ($row->{result_type} eq "result_formula")
              {
                if ($row->{underline} == 1 && $withUnderlines)
                {
                  $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
                }
                else
                {
                  $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
                }
              }
              else
              {
                $doc->print("&nbsp;");
              }
            }
          }
          if ($rowBold eq "1")
          {
            $doc->print("</b>");
          }
          $doc->print("</font></td>\n");
        }
        $doc->print("    </tr>\n");
      }
    }
    else
    {
      sub sort_name_num
      {
        my %account_hash = %{ $_[0] };

        if ($account_format == 0 || $account_format == 2)
        {
          return $account_hash{$a}{'cell_text'} cmp $account_hash{$b}{'cell_text'};
        }
        elsif ($account_format == 1 || $account_format == 3)
        {
          return $a <=> $b;
          #!!! compare by number is no longer valid.  change to use code
#          return $account_hash{$a}{'code'} cmp $account_hash{$b}{'code'};
        }
      }

      if (exists $row->{'error'} && $row->{'error'} == 1)
      {
        $error_occurred = 1;
      }

      if ($rowVisible eq "true")
      {
        $doc->print("    <tr>\n");
        $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
        for (my $skip_count=0; $skip_count < $row->{'display_level'}; $skip_count++)
        {
          $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
        }
        if (defined $row->{'sub_rows'} && $with_links eq "true")
        {
          my $open_close;
          if ($row->{'sub_rows_visible'} eq "true")
          {
            $open_close = "close";
          }
          else
          {
            $open_close = "open";
          }
          $doc->print('<span onclick="'."expandCollapse('$open_close', '$row_count', '-1'); document.run_form.submit();".'">');
          $doc->print('<img src="'.$self->{imageDir}.'/');
          if ($row->{'sub_rows_visible'} eq "true")
          {
            $doc->print('minus.gif" title="Collapse"');
          }
          else
          {
            $doc->print('plus.gif" title="Expand"');
          }
          $doc->print(' border="0"></span>&nbsp;');
        }
        if ($rowBold eq "1")
        {
          $doc->print("<b>");
        }
        my $hide_number = 0;
        if (exists $hide_numbers{$cell_type}{'0'})
        {
          $hide_number = 1;
        }
        $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $row->{'cell_text'}, code => $row->{'item_name'}));
        if (exists $row->{'error'} && $row->{'error'} == 1)
        {
          $doc->print("<font color=\"red\"><sup>+</sup></font>");
        }
        if ($rowBold eq "1")
        {
          $doc->print("</b>");
        }
        $doc->print("</font></td>\n");

        # first sub-level sub_row
        if (defined $row->{'sub_rows'} && $row->{'sub_rows_visible'} eq "true")
        {
          $doc->print('      <td class="numbers" colspan="' . (($args{'total_column'} eq "true" ? 1 : 0) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1)))) . '"><br></td>'."\n");
          foreach my $sub_count1 (sort { sort_name_num($row->{'sub_rows'}) } keys %{$row->{'sub_rows'}})  #!!! change out the keys to code instead of number?
          {
            my $subRow1 = $row->{sub_rows}{$sub_count1};
            next if ($subRow1->{visible} eq "false");
            $doc->print("    <tr>\n");
            $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
            for (my $skip_count=0; $skip_count < $subRow1->{'display_level'}; $skip_count++)
            {
              $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
            }
            if (defined $subRow1->{'sub_rows'} && $with_links eq "true")
            {
              my $open_close;
              if ($subRow1->{'sub_rows_visible'} eq "true")
              {
                $open_close = "close";
              }
              else
              {
                $open_close = "open";
              }
              $doc->print('<span onclick="'."expandCollapse('$open_close', '$row_count', '$sub_count1'); document.run_form.submit();".'">');
              $doc->print('<img src="'.$self->{imageDir}.'/');
              if ($subRow1->{'sub_rows_visible'} eq "true")
              {
                $doc->print('minus.gif" title="Collapse"');
              }
              else
              {
                $doc->print('plus.gif" title="Expand"');
              }
              $doc->print(' border="0"></a>&nbsp;');
            }
            if ($rowBold eq "1")
            {
              $doc->print("<b>");
            }
            my $hide_number = 0;
            if (exists $hide_numbers{$cell_type}{'1'})
            {
              $hide_number = 1;
            }
            $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $subRow1->{'cell_text'}, code => $subRow1->{'item_name'}));
            if ($rowBold eq "1")
            {
              $doc->print("</b>");
            }
            $doc->print("</font></td>\n");

            # second sub-level sub_row
            if (defined $subRow1->{'sub_rows'} && $subRow1->{'sub_rows_visible'} eq "true")
            {
              $doc->print('      <td class="numbers" colspan="' . (($args{'total_column'} eq "true" ? 1 : 0) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1)))) . "\"><br></td>\n");
              foreach my $sub_count2 (sort { sort_name_num($subRow1->{'sub_rows'}) } keys %{$subRow1->{'sub_rows'}})
              {
                my $subRow2 = $subRow1->{'sub_rows'}{$sub_count2};
                next if ($subRow2->{visible} eq "false");
                $doc->print("    <tr>\n");
                $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
                for (my $skip_count=0; $skip_count < $subRow2->{'display_level'}; $skip_count++)
                {
                  $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                }
                if (defined $subRow2->{'sub_rows'} && $with_links eq "true")
                {
                  my $open_close;
                  if ($subRow2->{'sub_rows_visible'} eq "true")
                  {
                    $open_close = "close";
                  }
                  else
                  {
                    $open_close = "open";
                  }
                  $doc->print('<span onclick="'."expandCollapse('$open_close', '$row_count', '$sub_count1|$sub_count2'); document.run_form.submit();".'">');
                  $doc->print('<img src="'.$self->{imageDir}.'/');
                  if ($subRow2->{'sub_rows_visible'} eq "true")
                  {
                    $doc->print('minus.gif" title="Collapse"');
                  }
                  else
                  {
                    $doc->print('plus.gif" title="Expand"');
                  }
                  $doc->print(' border="0"></a>&nbsp;');
                }
                if ($rowBold eq "1")
                {
                  $doc->print("<b>");
                }
                my $hide_number = 0;
                if (exists $hide_numbers{$cell_type}{'2'})
                {
                  $hide_number = 1;
                }
                $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $subRow2->{'cell_text'}, code => $subRow2->{'item_name'}));
                if ($rowBold eq "1")
                {
                  $doc->print("</b>");
                }
                $doc->print("</font></td>\n");

                # third sub-level sub_row
                if (defined $subRow2->{'sub_rows'} && $subRow2->{'sub_rows_visible'} eq "true")
                {
                  $doc->print('      <td class="numbers" colspan="' . (($args{'total_column'} eq "true" ? 1 : 0) + ((scalar @{$args{'dates_entered'}}) * ((scalar @column_data > 0 ? scalar @column_data : 1)))) . "\"><br></td>\n");
                  foreach my $sub_count3 (sort { sort_name_num($subRow2->{'sub_rows'}) } keys %{$subRow2->{'sub_rows'}})
                  {
                    my $subRow3 = $subRow2->{sub_rows}{$sub_count3};
                    next if ($subRow3->{visible} eq "false");
                    $doc->print("    <tr>\n");
                    $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
                    for (my $skip_count=0; $skip_count < $subRow3->{'display_level'}; $skip_count++)
                    {
                      $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                    }
                    if (defined $subRow3->{'sub_rows'} && $with_links eq "true")
                    {
                      my $open_close;
                      if ($subRow3->{'sub_rows_visible'} eq "true")
                      {
                        $open_close = "close";
                      }
                      else
                      {
                        $open_close = "open";
                      }
                      $doc->print('<span onclick="'."expandCollapse('$open_close', '$row_count', '$sub_count1|$sub_count2|$sub_count3'); document.run_form.submit();".'">');
                      $doc->print('<img src="'.$self->{imageDir}.'/');
                      if ($subRow3->{'sub_rows_visible'} eq "true")
                      {
                        $doc->print('minus.gif" title="Collapse"');
                      }
                      else
                      {
                        $doc->print('plus.gif" title="Expand"');
                      }
                      $doc->print(' border="0"></a>&nbsp;');
                    }
                    if ($rowBold eq "1")
                    {
                      $doc->print("<b>");
                    }
                    my $hide_number = 0;
                    if (exists $hide_numbers{$cell_type}{'3'})
                    {
                      $hide_number = 1;
                    }
                    $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $subRow3->{'cell_text'}, code => $subRow3->{'item_name'}));
                    if ($rowBold eq "1")
                    {
                      $doc->print("</b>");
                    }
                    $doc->print("</font></td>\n");
                    for (my $value_count=0; $value_count < scalar @{$subRow3->{'values'}}; $value_count++)
                    {
                      my $valueCol = $subRow3->{values}[$value_count];
                      for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
                      {
                        $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow3->{'display_level'} * 1.5) . 'em;">');
                        if ($rowDrillable && $subRow3->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                        {
                          (my $tmp_sql_clause = $subRow3->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                          my $hide_number = 0;
                          if (exists $hide_numbers{$cell_type}{'3'})
                          {
                            $hide_number = 1;
                          }
                          $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $subRow3->{cell_text}', '$tmp_sql_clause');">
ENDOF
                        }
                        $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
                        if ($rowBold eq "1")
                        {
                          $doc->print("<b>");
                        }
                        if ($valueCol->[$value_count2] < 0)
                        {
                          $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                        }
                        else
                        {
                          $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                        }
                        if ($rowBold eq "1")
                        {
                          $doc->print("</b>");
                        }
                        if (exists $subRow3->{'error'} && $subRow3->{'error'} == 1)
                        {
                          $doc->print("<font color=\"red\"><sup>*</sup></font>");
                        }
                        $doc->print("</font>");
                        if ($rowDrillable && $subRow3->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                        {
                          $doc->print("</a>");
                        }
                        $doc->print("</td>\n");
                      }
                    }
                    if ($args{'total_column'} eq "true")
                    {
                      $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow3->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
                      if ($rowBold eq "1")
                      {
                        $doc->print("<b>");
                      }
                      if ($subRow3->{'total'} < 0)
                      {
                        $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow3->{'total'})))));
                      }
                      else
                      {
                        $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow3->{'total'})))));
                      }
                      if ($rowBold eq "1")
                      {
                        $doc->print("</b>");
                      }
                      $doc->print("</font></td>\n");
                    }
                    $doc->print("    </tr>\n");
                  }

                  $doc->print("    <tr>\n");
                  $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
                  for (my $skip_count=0; $skip_count < $subRow2->{'display_level'}; $skip_count++)
                  {
                    $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                  }
                  if ($rowBold eq "1")
                  {
                    $doc->print("<b>");
                  }
                  my $hide_number = 0;
                  if (exists $hide_numbers{$cell_type}{'2'})
                  {
                    $hide_number = 1;
                  }
                  $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $subRow2->{'cell_text'}, code => $subRow2->{'item_name'}));
                  if ($rowBold eq "1")
                  {
                    $doc->print("</b>");
                  }
                  $doc->print("&nbsp;Total:");
                  $doc->print("</font></td>\n");
                  for (my $value_count=0; $value_count < scalar @{$subRow2->{'values'}}; $value_count++)
                  {
                    my $valueCol = $subRow2->{values}[$value_count];
                    for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
                    {
                      $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow2->{'display_level'} * 1.5) . "em;\">");
                      if ($rowDrillable && $subRow2->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                      {
                        (my $tmp_sql_clause = $subRow2->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                        my $hide_number = 0;
                        if (exists $hide_numbers{$cell_type}{'2'})
                        {
                          $hide_number = 1;
                        }
                        $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $subRow2->{cell_text}', $tmp_sql_clause);">
ENDOF
                      }
                      $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
                      if ($rowBold eq "1")
                      {
                        $doc->print("<b>");
                      }
                      if ($valueCol->[$value_count2] < 0)
                      {  # putting this back to format_value
                        $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                      }
                      else
                      {  # putting this back to format_value
                        $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                      }
                      if ($rowBold eq "1")
                      {
                        $doc->print("</b>");
                      }
                      $doc->print("</font>");
                      if ($rowDrillable && $subRow2->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                      {
                        $doc->print("</a>");
                      }
                      $doc->print("</font></td>\n");
                    }
                  }
                  if ($args{'total_column'} eq "true")
                  {
                    $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow2->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
                    if ($rowBold eq "1")
                    {
                      $doc->print("<b>");
                    }
                    if ($subRow2->{'total'} < 0)
                    {
                      $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow2->{'total'})))));
                    }
                    else
                    {
                      $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow2->{'total'})))));
                    }
                    if ($rowBold eq "1")
                    {
                      $doc->print("</b>");
                    }
                    $doc->print("</font></td>\n");
                  }
                  $doc->print("    </tr>\n");
                } # end third sub-level sub_row
                else
                {
                  for (my $value_count=0; $value_count < scalar @{$subRow2->{'values'}}; $value_count++)
                  {
                    my $valueCol = $subRow2->{values}[$value_count];
                    for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
                    {
                      $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow2->{'display_level'} * 1.5) . 'em;">');
                      if ($rowDrillable && $subRow2->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                      {
                        (my $tmp_sql_clause = $subRow2->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                        my $hide_number = 0;
                        if (exists $hide_numbers{$cell_type}{'2'})
                        {
                          $hide_number = 1;
                        }
                        $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $subRow2->{cell_text}', '$tmp_sql_clause');">
ENDOF
                      }
                      $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
                      if ($rowBold eq "1")
                      {
                        $doc->print("<b>");
                      }
                      if ($valueCol->[$value_count2] < 0)
                      {
                        $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                      }
                      else
                      {
                        $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                      }
                      if ($rowBold eq "1")
                      {
                        $doc->print("</b>");
                      }
                      if (exists $subRow2->{'error'} && $subRow2->{'error'} == 1)
                      {
                        $doc->print("<font color=\"red\"><sup>*</sup></font>");
                      }
                      $doc->print("</font>");
                      if ($rowDrillable && $subRow2->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                      {
                        $doc->print("</a>");
                      }
                      $doc->print("</td>\n");
                    }
                  }
                  if ($args{'total_column'} eq "true")
                  {
                    $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow2->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
                    if ($rowBold eq "1")
                    {
                      $doc->print("<b>");
                    }
                    if ($subRow2->{'total'} < 0)
                    {
                      $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow2->{'total'})))));
                    }
                    else
                    {
                      $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow2->{'total'})))));
                    }
                    if ($rowBold eq "1")
                    {
                      $doc->print("</b>");
                    }
                    $doc->print("</font></td>\n");
                  }
                  $doc->print("    </tr>\n");
                }
              }

              $doc->print("    <tr>\n");
              $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
              for (my $skip_count=0; $skip_count < $subRow1->{'display_level'}; $skip_count++)
              {
                $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
              }
              if ($rowBold eq "1")
              {
                $doc->print("<b>");
              }
              my $hide_number = 0;
              if (exists $hide_numbers{$cell_type}{'1'})
              {
                $hide_number = 1;
              }
              $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $subRow1->{'cell_text'}, code => $subRow1->{'item_name'}));
              if ($rowBold eq "1")
              {
                $doc->print("</b>");
              }
              $doc->print("&nbsp;Total:");
              $doc->print("</font></td>\n");
              for (my $value_count=0; $value_count < scalar @{$subRow1->{'values'}}; $value_count++)
              {
                my $valueCol = $subRow1->{values}[$value_count];
                for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
                {
                  $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow1->{'display_level'} * 1.5) . 'em;">');
                  if ($rowDrillable && $subRow1->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                  {
                    (my $tmp_sql_clause = $subRow1->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                    my $hide_number = 0;
                    if (exists $hide_numbers{$cell_type}{'1'})
                    {
                      $hide_number = 1;
                    }
                    $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $subRow1->{cell_text}', '$tmp_sql_clause');">
ENDOF
                  }
                  $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
                  if ($rowBold eq "1")
                  {
                    $doc->print("<b>");
                  }
                  if ($withUnderlines)
                  {
                    if ($valueCol->[$value_count2] < 0)
                    {
                      $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                    }
                    else
                    {
                      $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                    }
                  }
                  else
                  {
                    if ($valueCol->[$value_count2] < 0)
                    {
                      $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                    }
                    else
                    {
                      $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                    }
                  }
                  if ($rowBold eq "1")
                  {
                    $doc->print("</b>");
                  }
                  $doc->print("</font>");
                  if ($subRow1->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                  {
                    $doc->print("</a>");
                  }
                  $doc->print("</td>\n");
                }
              }
              if ($args{'total_column'} eq "true")
              {
                $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow1->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
                if ($rowBold eq "1")
                {
                  $doc->print("<b>");
                }
                if ($withUnderlines)
                {
                  if ($subRow1->{'total'} < 0)
                  {
                    $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow1->{'total'})))));
                  }
                  else
                  {
                    $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow1->{'total'})))));
                  }
                }
                else
                {
                  if ($subRow1->{'total'} < 0)
                  {
                    $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow1->{'total'})))));
                  }
                  else
                  {
                    $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow1->{'total'})))));
                  }
                }
                if ($rowBold eq "1")
                {
                  $doc->print("</b>");
                }
                $doc->print("</font></td>\n");
              }
              $doc->print("    </tr>\n");
            } # end second sub-level sub_row
            else
            {
              for (my $value_count=0; $value_count < scalar @{$subRow1->{'values'}}; $value_count++)
              {
                my $valueCol = $subRow1->{values}[$value_count];
                for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
                {
                  $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow1->{'display_level'} * 1.5) . 'em;">');
                  if ($rowDrillable && $subRow1->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                  {
                    (my $tmp_sql_clause = $subRow1->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                    my $hide_number = 0;
                    if (exists $hide_numbers{$cell_type}{'1'})
                    {
                      $hide_number = 1;
                    }
                    $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $subRow1->{cell_text}', '$tmp_sql_clause');">
ENDOF
                  }
                  $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
                  if ($rowBold eq "1")
                  {
                    $doc->print("<b>");
                  }
                  if ($valueCol->[$value_count2] < 0)
                  {
                    $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                  }
                  else
                  {
                    $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                  }
                  if ($rowBold eq "1")
                  {
                    $doc->print("</b>");
                  }
                  if (exists $subRow1->{'error'} && $subRow1->{'error'} == 1)
                  {
                    $doc->print("<font color=\"red\"><sup>*</sup></font>");
                  }
                  $doc->print("</font>");
                  if ($rowDrillable && $subRow1->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
                  {
                    $doc->print("</a>");
                  }
                  $doc->print("</td>\n");
                }
              }
              if ($args{'total_column'} eq "true")
              {
                $doc->print('      <td class="numbers" style="padding-right: ' . ($subRow1->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
                if ($rowBold eq "1")
                {
                  $doc->print("<b>");
                }
                if ($subRow1->{'total'} < 0)
                {
                  $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow1->{'total'})))));
                }
                else
                {
                  $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($subRow1->{'total'})))));
                }
                if ($rowBold eq "1")
                {
                  $doc->print("</b>");
                }
                $doc->print("</font></td>\n");
              }
              $doc->print("    </tr>\n");
            }
          }

          $doc->print("    <tr>\n");
          $doc->print('      <td class="text" colspan="' . ($max_display_level + 1) . "\" nowrap><font size=\"$text_size\" face=\"$text_font\">");
          for (my $skip_count=0; $skip_count < $row->{'display_level'}; $skip_count++)
          {
            $doc->print("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
          }
          if ($rowBold eq "1")
          {
            $doc->print("<b>");
          }
          my $hide_number = 0;
          if (exists $hide_numbers{$cell_type}{'0'})
          {
            $hide_number = 1;
          }
          $doc->print($self->format_account(hide_number => $hide_number, account_format => $account_format, name => $row->{'cell_text'}, code => $row->{'item_name'}));
          if ($rowBold eq "1")
          {
            $doc->print("</b>");
          }
          $doc->print("&nbsp;Total:");
          $doc->print("</font></td>\n");
          for (my $value_count=0; $value_count < scalar @{$row->{'values'}}; $value_count++)
          {
            my $valueCol = $row->{values}[$value_count];
            for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
            {
              $doc->print('      <td class="numbers" style="padding-right: ' . ($row->{'display_level'} * 1.5) . 'em;">');
              if ($rowDrillable && $row->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
              {
                (my $tmp_sql_clause = $row->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                my $hide_number = 0;
                if (exists $hide_numbers{$cell_type}{'0'})
                {
                  $hide_number = 1;
                }
                $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $row->{cell_text}', '$tmp_sql_clause');">
ENDOF
              }
              $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
              if ($rowBold eq "1")
              {
                $doc->print("<b>");
              }
              if ($withUnderlines)
              {
                if ($valueCol->[$value_count2] < 0)
                {
                  $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                }
                else
                {
                  $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                }
              }
              else
              {
                if ($valueCol->[$value_count2] < 0)
                {
                  $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                }
                else
                {
                  $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
                }
              }
              if ($rowBold eq "1")
              {
                $doc->print("</b>");
              }
              $doc->print("</font>");
              if ($rowDrillable && $row->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
              {
                $doc->print("</a>");
              }
              $doc->print("</td>\n");
            }
          }
          if ($args{'total_column'} eq "true")
          {
            $doc->print('      <td class="numbers" style="padding-right: ' . ($row->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
            if ($rowBold eq "1")
            {
              $doc->print("<b>");
            }
            if ($withUnderlines)
            {
              if ($rowTotal < 0)
              {
                $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
              }
              else
              {
                $doc->print($self->formatUnderlineValue(underlineLength => $underline_length, value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
              }
            }
            else
            {
              if ($rowTotal < 0)
              {
                $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
              }
              else
              {
                $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
              }
            }
            if ($rowBold eq "1")
            {
              $doc->print("</b>");
            }
            $doc->print("</font></td>\n");
          }
          $doc->print("    </tr>\n");
        } # end first sub-level sub_row
        else
        {
          for (my $value_count=0; $value_count < scalar @{$row->{'values'}}; $value_count++)
          {
            my $valueCol = $row->{values}[$value_count];
            for (my $value_count2=0; $value_count2 < scalar @{$valueCol}; $value_count2++)
            {
              $doc->print('      <td class="numbers" style="padding-right: ' . ($row->{'display_level'} * 1.5) . 'em;">');
              if ($rowDrillable && $row->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
              {
                (my $tmp_sql_clause = $row->{'sql_clause'}[$value_count][$value_count2]) =~ s/$singleQuoteRE/\\\'/g;
                my $hide_number = 0;
                if (exists $hide_numbers{$cell_type}{'0'})
                {
                  $hide_number = 1;
                }
                $doc->print(<<"ENDOF");
<a href="javascript: drill_value('$args{report_name} - $row->{cell_text}', '$tmp_sql_clause');">
ENDOF
              }
              $doc->print("<font size=\"$data_size\" face=\"$data_font\">");
              if ($rowBold eq "1")
              {
                $doc->print("<b>");
              }
              if ($valueCol->[$value_count2] < 0)
              {
                $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
              }
              else
              {
                $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($valueCol->[$value_count2])))));
              }
              if ($rowBold eq "1")
              {
                $doc->print("</b>");
              }
              if (exists $row->{'error'} && $row->{'error'} == 1)
              {
                $doc->print("<font color=\"red\"><sup>*</sup></font>");
              }
              $doc->print("</font>");
              if ($rowDrillable && $row->{'sql_clause'}[$value_count][$value_count2] ne "invalid" && $with_links eq "true")
              {
                $doc->print("</a>");
              }
              $doc->print("</td>\n");
            }
          }
          if ($args{'total_column'} eq "true")
          {
            $doc->print('      <td class="numbers" style="padding-right: ' . ($row->{'display_level'} * 1.5) . "em;\"><font size=\"$data_size\" face=\"$data_font\">");
            if ($rowBold eq "1")
            {
              $doc->print("<b>");
            }
            if ($rowTotal < 0)
            {
              $doc->print($self->format_value(negative_format => $negative_format, sign => "-", value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
            }
            else
            {
              $doc->print($self->format_value(value => $self->commify($self->massage_data(precision => $args{'data_precision'}, value => abs($rowTotal)))));
            }
            if ($rowBold eq "1")
            {
              $doc->print("</b>");
            }
            $doc->print("</font></td>\n");
          }
          $doc->print("    </tr>\n");
        }
      }
    }
  }

  $doc->print(<<"END_OF_BODY");
  </table>
  </form>
</div>
END_OF_BODY
  if ($error_occurred)
  {
    $doc->print("<font color=\"red\"><b><sup>+</sup>:&nbsp;&nbsp;An error occurred while processing the report for this row.</b></font><br>\n");
    $doc->print("<font color=\"red\"><b><sup>*</sup>:&nbsp;&nbsp;This value contains invalid data.&nbsp;&nbsp;Most likely, the column modifier is not allowed or invalid.</b></font>\n");
  }

  if ($input{'report_command'} eq "print")
  {
    my $onload_code = <<"END_OF_ONLOAD";
    if ((navigator.appName.indexOf("Microsoft") != -1 && parseInt(navigator.appVersion, 10) <= 4) || parseInt(navigator.appVersion < 4))
    {
      alert("This browser does not support automatic printing of the document.\\nPlease print the document manually.");
    }
    else
    {
      window.print();
    }
END_OF_ONLOAD

    $doc->setOnload( code => $onload_code);
  }
  else
  {
    my $onloadCode = <<"END_OF_ONLOAD";
window.scrollTo(0, $args{pageTop});
END_OF_ONLOAD

    $doc->setOnload(code => $onloadCode);
  }

  return $doc;
}
    
#****************************** Helper functions *****************************

###################### Data Gathering Functions converted from reports.pm to DBI structure #############################

# massage_data
# takes  : precision (defaults to 1), value
# returns: value (massaged into the specified precision).
sub massage_data
{
  my $self = shift;
  my %arguments = ( precision => 1, @_, );
  my $precision = $arguments{'precision'};
  my $value = $arguments{'value'};

  if (length $value == 0)
  {
    # default to returning a zero value
    $value = 0.00;
  }

  if ($precision == 1)
  {
    $value = sprintf "%.2f", $value;
  }
  elsif ($precision < 0)
  {  # handle the special case where the user wants X number of decimal points other than 2.
    $precision =~ s/^-//;  # remove the -.
    $value = sprintf "%.*f", $precision, $value;
  }
  else
  {
    $value /= $precision;  # do the precision modification
    $value = sprintf "%.0f", $value;  # do any rounding needed
    $value = int($value);  # cast it to a whole number
  }

  return $value;
}

sub validateParameter
{
  my $self = shift;
  # this function takes the name of the parameter and it's value and does the work of making sure that the value is valid for that parameter.
  my ($name, $value) = ( @_ );
  my $result = 1;

  if (length $name == 0)
  {
    $result = 0;
  }
  if (length $value == 0)
  {
    $result = 0;
  }

  if ($name eq "mode")
  {
    if ($value ne "edit" && $value ne "create")
    {
      $result = 0;
    }
  }
  elsif ($name eq "name")
  {
  }
  elsif ($name eq "original_name")
  {
  }
  elsif ($name eq "id")
  {
    if ($value < -1)
    {
      $result = 0;
    }
  }
  elsif ($name eq "report_class")
  {
    if (!exists $report_classes{$value})
    {
      $result = 0;
    }
  }
  elsif ($name eq "report_type")
  {
    if ($value ne "r" && $value ne "d")
    {
      $result = 0;
    }
  }
  elsif ($name eq "column_variable")
  {
  }
  elsif ($name eq "column_variable_data")
  {
  }
  elsif ($name eq "date_range")
  {
    if ($value ne "true" && $value ne "false")
    {
      $result = 0;
    }
  }
  elsif ($name eq "display_dates")
  {
    if ($value ne "true" && $value ne "false")
    {
      $result = 0;
    }
  }
  elsif ($name eq "max_columns")
  {
  }
  elsif ($name eq "min_columns")
  {
  }
  elsif ($name eq "text_size")
  {
  }
  elsif ($name eq "text_font")
  {
  }
  elsif ($name eq "data_size")
  {
  }
  elsif ($name eq "data_font")
  {
  }
  elsif ($name eq "total_column")
  {
  }
  elsif ($name eq "underline_length")
  {
  }
  elsif ($name eq "error_string")
  {
  }
  elsif ($name eq "rows")
  {
    my @rows = split /\n/, $value;
    if (scalar @rows == 0)
    {
      $result = 0;
    }
  }
  elsif ($name eq "file_name")
  {
  }
  elsif ($name eq "report_group")
  {
    if ($value ne "regular")
    {
      $result = 0;
    }
  }
  elsif ($name eq "new_report_name")
  {
  }
  elsif ($name eq "windowName")
  {
    $result = 0 if ($value !~ /^(edit_report_window)$/);
  }
  else
  {
    $result = 0;  # signal that an error occurred as we do not know anything about this variable.
  }

  return $result;
}

# display_regular_report_top
# takes  : filename
# returns: nothing
sub display_regular_report_top
{
  my $self = shift;
  my %args = ( @_, );
  my $file_name = $self->{input}->{'file_name'};
  my $count;
  my $args_ref;
  my @data = ();

  $args_ref = retrieve($file_name);
  %args = %{@{$args_ref}[0]};
  @data = @{@{$args_ref}[1]};
  my $id = $args{'id'};
  my $ledgerCode = $self->{input}->{'ledgerCode'};
  my $max_display_level = $args{'max_display_level'};

  my $doc = $self->prepDoc;
  my $onload_code = <<"END_OF_ONLOAD";
    var bottomWindowName = '_b';
    bottom_win_ref = window.open("", bottomWindowName);
END_OF_ONLOAD
  $doc->setOnload(code => $onload_code);

  $doc->setFocus("body");
  $doc->setLinkDecorations();
  # set the overall font size to be smaller
  $doc->setStyleEntry(tag => "body", string => "font-size: .5em;");

  my $dates_entered = join ("|", @{$args{'dates_entered'}});

#converted js customize function to hidden div.
  $doc->print(css => [ "textarea,select,form,input { font-size: inherit; }" ]);
  $doc->print(css => [ "input { border: none; border-style: solid; border-width: 1px; padding: .1em; }" ]);
  $doc->print(css => [ "div { padding: .5em; }" ]);
  my $dataFont = <<BODY;
  <div style="float: right;">
    Data&nbsp;Font&nbsp;Size: <select name="data_size" onChange="data_size_change(this);">
BODY
  for ($count=0; $count < scalar @font_sizes; $count++)
  {
    $dataFont .= "      <option value=\"$font_sizes[$count]\">$font_sizes{$font_sizes[$count]}\n";
  }
  $dataFont .= <<"BODY";
    </select>
  </div>
BODY
  my $textFont = <<BODY;
  <div style="float: right;">
    Text&nbsp;Font&nbsp;Size: <select name="text_size" onChange="text_size_change(this);">
BODY
  for ($count=0; $count < scalar @font_sizes; $count++)
  {
    $textFont .= "      <option value=\"$font_sizes[$count]\">$font_sizes{$font_sizes[$count]}</option>\n";
  }
  $textFont .= <<"BODY";
    </select>
  </div>
BODY
  my $negative_formats = <<"BODY";
    <div style="float: right;">
    Negative&nbsp;Value&nbsp;Format:
BODY
  $negative_formats .= "    <input type=radio name=\"negative_format\" value=\"-1\" onclick=\"negative_format_click(this);\" selected=\"1\">Default\n";
  for ($count=0; $count < scalar @negative_formats; $count++)
  {
    $negative_formats .= "    <input type=radio name=\"negative_format\" value=\"$negative_formats[$count]\" onclick=\"negative_format_click(this);\">&nbsp;$negative_formats{$negative_formats[$count]}\n";
  }
  $negative_formats .= <<"BODY";
    </div>
BODY
  my $under = <<"BODY";  
  <div style="float: right;">
    Underline Length:
    <select name="underline_length" onChange="underline_length_change(this);">
BODY
  for ($count=$self->{min_underline_length}; $count <= $self->{max_underline_length}; $count++)
  {
    $under .= "        <option value=\"$count\">$count</option>\n";
  }
  $under .= <<"BODY";
      </select>
    </div>
BODY
  my $total_column = '';
  if ($args{report_class} ne 'balance')
  {
    $total_column = <<"FOO";
    <div style="float: left;">
      Total Column:
      <select name="total_column" onChange="total_column_change(this);">
        <option value="$t_or_f{'0'}">No</option>
        <option value="$t_or_f{'1'}">Yes</option>
      </select>
    </div>
FOO
  }

  #start output of main form  
  $doc->print(<<"BODY");
<form method="post" name="run_form" target="_top" class="primary">
$self->{postInfo}
  <input type="hidden" name="command" value="" />
  <input type="hidden" name="id" value="$id" />
  <input type="hidden" name="file_name" value="$file_name" />
  <div id="main" style="text-align: center;">
    <a href="javascript: display_report();">Re-display</a>
    | <a href="javascript: print_report();">Print</a>
    | <a href="javascript: document.run_form.reset();">Reset</a>
    | <a href="javascript: rerun_report(document.run_form);">Re-run</a>
    | <a href="javascript: expand_levels();">Show&nbsp;Level(s)</a>
    &nbsp;<select name="levels_select" onChange="expand_levels();">
      <option value="-1">All
      <option value="-2">Smart
BODY
  for ($count=0; $count < $max_display_level + 1; $count++)
  {
    $doc->print("          <option value=\"$count\">$column_names{($count + 1)}\n");
  }
  $doc->print(<<"BODY");
    </select>
    | <a href="javascript: export_report();">Export</a>
  </div>

  <div id="customizeDiv">
    <div style="float: left;">
      Report&nbsp;Name:<input type=text name="report_name" value="$args{report_name}" size="30" onChange="report_name_change(this);">
    </div>
    <div style="float: left;">
      Printed Report Width: <input type="text" name="reportWidth" size="3" maxlength="3" value="100" onChange="reportWidthChange(this);">%
    </div>
    $textFont
    $total_column
    $dataFont
    <div style="float: left;">
      Border: <input type=checkbox onclick="border_change(this);">
    </div>
    <div style="float: left;">
      Display Zero Rows? <input type=checkbox onclick="displayZeroRows_change(this);">
    </div>
    $under
    $negative_formats
    <div style="display: none;">
      Labels&nbsp;as:  <select name="account_format" onChange="account_format_change(this);">
BODY
  foreach my $key (sort keys %{$self->{account_formats}} )
  {
    my $selected = '';
    if ($key eq '2') { $selected = ' selected="1" '; }
    $doc->print('            <option value="'.$key.'"'.$selected.'>'.$self->{account_formats}->{$key}."</option>\n");
  }
  $doc->print(<<"BODY");
      </select>
    </div>
  </div>
</form>
BODY

  return $doc;
}

# bool areAllColumnsZero(row)
#
# returns 1 if all the columns in the row are "empty",
# else returns 0.
# will return 0 if the total is not 0.
sub areAllColumnsZero
{
  my $self = shift;
  my $row = shift;
  # immediate bail out check.
  return 0 if ($row->{total} != 0);

#  return $row->{allColumnsZero} if (exists $row->{allColumnsZero});

  #initialize shortcut to true
  $row->{allColumnsZero} = 1;
#  my $nonEmptyRows = 0;
#  my $emptyRows = 0;
  if (exists $row->{values})
  {
    for (my $column=0; $column < scalar @{$row->{values}}; $column++)
    {
      my $valueCol = $row->{values}[$column];
      for (my $column2=0; $column2 < scalar @{$valueCol}; $column2++)
      {
#         $nonEmptyRows++ if ($valueCol->[$column2] != 0);
#         $emptyRows++ if ($valueCol->[$column2] == 0);
        #if finding any data turn set shortcut to false
        if (sprintf("%.2f",$valueCol->[$column2]) != 0.00) { $row->{allColumnsZero} = 0; last; }
      }
    }
  }

#  $row->{allColumnsZero} = 0 if ($nonEmptyRows > 0);
#   $row->{allColumnsZero} = 1 if ($nonEmptyRows == 0 && $emptyRows == 0);
#   $row->{allColumnsZero} = 1 if ($nonEmptyRows == 0 && $emptyRows > 0);
#   $row->{allColumnsZero} = 0 if ($nonEmptyRows > 0 && $emptyRows >= 0);
  #$row->{allColumnsZero} = 1 if ($row->{total} == 0);  # catchall case.

  return $row->{allColumnsZero};
}

# bool $self->updateRowVisibility(row, displayZeroRows, cellType)
# determines if the given row should be visible or not
# and any directly descendant sub-rows.
# if we return 1, then continue processing,
# if we return 0, then you are not going to be seeing this row.
sub updateRowVisibility
{
  my $self = shift;
  my $row = shift;
  my $displayZeroRows = shift;
  my $cellType = shift;
  my $total = $row->{'total'};

  if ($cellType eq 'total' || $cellType eq 'compare' || $cellType eq 'text' || $cellType eq 'line')
  {
    return 1;
  }
  if (!$displayZeroRows && $total == 0.00)
  {
    # only set the row hidden, if all the columns are = 0.
    if ($self->areAllColumnsZero($row))
    {
      $row->{visible} = "false";
      return 0;
    }
  }
  return 1;
}

# bool $self->updateSubRowVisibility(subRow, displayZeroRows, cellType)
# determines if the given sub-row should be visible or not.
# if we return 1, then continue processing,
# if we return 0, then you are not going to be seeing this row.
sub updateSubRowVisibility
{
  my $self = shift;
  return $self->updateRowVisibility(@_);
}

# void $self->expandRow(row, displayZeroRows, level, subRow, debugStr)
# expand the specified row and take into account the displayZeroRows value
# if the user specifies subRow (which is an anonymous array) then we only expand
# the specified subRow values.  format of subRow is [ subRow, subRow, subRow ]
# where each entry is the value to expand under the previous entry.
# ex. row = Acct Receivable, subRow = [ 1500 ] would expand just the AR subrow.
# row = Current Liabilities, subRow = [ 2100, 4456 ] would attempt to expand
#     the AP then HLR sub-rows.
# If subRow is empty, then we default to expand all subrows under the row we
# are working with.
sub expandRow
{
  my $self = shift;
  my $row = shift;
  my $displayZeroRows = shift;
  my $level = shift;
  my $subRow = shift;
  my $debugStr = shift;

  my $depth = $row->{'display_level'};
  my $cellType = $row->{'cell_type'};
  my $total = $row->{'total'};

  #${$debugStr} .= "cellType = '$cellType', depth = '$depth', total = '$total', level = '$level', subRow = [" . join(",", @{$subRow}) . "]<br>\n";

  # always show the base level for the row; a defined row of the report will always be shown no matter what its level
  # unless it has been defined as an invisible row
  # level = -2 will implement the Smart opening code.
  if ($row->{'show'} eq "1")
  {
    $row->{'visible'} = "true";
  }
  else
  {
    $row->{'visible'} = "false";
  }
  if (!$self->updateRowVisibility($row, $displayZeroRows, $cellType))
  {
    return;
  }
  if ($level =~ /^(-2|-3)$/)
  {
    my $allColumnsZero = $self->areAllColumnsZero($row);
    if ($cellType =~ /$tripleAccountsRE/ && !$allColumnsZero)
    {
      if (defined $row->{'sub_rows'})
      {
        # make sure that we open this row, even though all the entries under it may
        # not be openable!
        $row->{'sub_rows_visible'} = "true";
      }
      else
      {
        $row->{'sub_rows_visible'} = "false";
      }
    }
  }
  elsif ($level =~ /^(-4)$/)
  {
    if ($cellType =~ /$tripleAccountsRE/ && $total == 0)
    {
      if (!$displayZeroRows)
      {
        $row->{visible} = "false";
      }
      return;
    }
  }
  if (defined $row->{'sub_rows'})
  {
    foreach my $subCount (keys %{$row->{'sub_rows'}})
    {
      if (scalar @{$subRow} == 0 || ($subRow->[0] eq $subCount))
      {
        #${$debugStr} .= "expandingSubRow = '$subCount'<br>\n";
        my $currSubRow = $row->{'sub_rows'}{$subCount};
        $self->expandSubRow($row, $currSubRow, $displayZeroRows, $level, $depth + 1, $cellType, $total, $subRow, $debugStr);
      }
    }
    # reloop over and make sure the sub_rows_visible flag is correct.
    my $subRowVisible = "false";
    foreach my $subCount (keys %{$row->{'sub_rows'}})
    {
      if ($row->{'sub_rows'}{$subCount}{'visible'} eq "true")
      {
        $subRowVisible = "true";
      }
    }
    if ($subRowVisible eq "false")
    {
      $row->{'sub_rows_visible'} = "false";
    }
  }
}

# void $self->expandSubRow(parentRow, currSubRow, displayZeroRows, level, depth, cellType, total, subRow, debugStr)
sub expandSubRow
{
  my $self = shift;
  my $parentRow = shift;
  my $currSubRow = shift;
  my $displayZeroRows = shift;
  my $level = shift;
  my $depth = shift;
  my $cellType = shift;
  my $total = shift;
  my $subRow = shift;
  my $debugStr = shift;

  my $displayChildren = 0;

  #${$debugStr} .= "&nbsp;&nbsp;&nbsp;&nbsp;subrow: $currSubRow->{item_name}";
  #foreach my $name (keys %{$currSubRow})
  #{
  #  ${$debugStr} .= ", $name:$currSubRow->{$name}";
  #}
  #${$debugStr} .= "<br>\n";

  #${$debugStr} .= "&nbsp;&nbsp;&nbsp; depth = '$depth', level = '$level', subRow = [" . join(",", @{$subRow}) . "]<br>\n";

  if (!$self->updateSubRowVisibility($currSubRow, $displayZeroRows, $cellType))
  {
    return;
  }

  if ($level =~ /$negativeOneRE/ || $level >= $depth)
  {
    $parentRow->{'sub_rows_visible'} = "true";
    $currSubRow->{'visible'} = "true";
  }
  elsif ($level eq "-2")
  {
    if ($cellType =~ /$tripleAccountsRE/)
    {
      my $allColumnsZero = $self->areAllColumnsZero($currSubRow);
      if (!$allColumnsZero)
      {
        if (defined $currSubRow->{'sub_rows'})
        {
          my $subTotal = $currSubRow->{'total'};
          my @items = keys %{$currSubRow->{'sub_rows'}};
          if ($subTotal != 0)
          {
            if (scalar @items > 1)
            {
              $parentRow->{'sub_rows_visible'} = "true";
              $currSubRow->{'visible'} = "true";
            }
            elsif (scalar @items == 1)
            {
              # if the sub account isn't -1, then open it.
              if ($items[0] ne "-1")
              {
                $parentRow->{'sub_rows_visible'} = "true";
                $currSubRow->{'visible'} = "true";
              }
              else
              {
                # make sure this sub_row is not opened.
                $currSubRow->{'sub_rows_visible'} = "false";
              }
            }
          }
          else
          {
            # make sure this sub_row is not opened.
            $currSubRow->{'sub_rows_visible'} = "false";
            if (!$displayZeroRows)
            {
              $currSubRow->{visible} = "false";
            }
          }
        }
        else
        {
          my $subTotal = $currSubRow->{'total'};
          my @items = keys %{$parentRow->{'sub_rows'}};
          if ($subTotal != 0)
          {
            if (scalar @items > 1)
            {
              $parentRow->{'sub_rows_visible'} = "true";
              $currSubRow->{'visible'} = "true";
              $displayChildren = 1;
            }
            elsif (scalar @items == 1)
            {
              # if the sub account isn't -1, then open it.
              if ($items[0] ne "-1")
              {
                $parentRow->{'sub_rows_visible'} = "true";
                $currSubRow->{'visible'} = "true";
                $displayChildren = 1;
              }
            }
          }
          else
          {
            if (!$displayZeroRows)
            {
              $currSubRow->{visible} = "false";
            }
          }
        }
      }
      else
      {
        # make sure it is closed.
        $currSubRow->{'sub_rows_visible'} = "false";
        if (!$displayZeroRows)
        {
          $currSubRow->{visible} = "false";
        }
      }
    }
    else
    {
      #$currSubRow->{'visible'} = "true";
    }
  }
  elsif ($level eq "-3")
  {
    $parentRow->{'sub_rows_visible'} = "true";
    $currSubRow->{visible} = "true";
    if (scalar @{$subRow} >= 1)
    {
      $currSubRow->{sub_rows_visible} = "true";
    }
    else
    {
      return;
    }
  }
  elsif ($level eq "-4")
  {
    if ($cellType =~ /$tripleAccountsRE/)
    {
      my $subTotal = $currSubRow->{'total'};
      if ($subTotal == 0)
      {
        if (!$displayZeroRows)
        {
          $currSubRow->{visible} = "false";
        }
        return;
      }
    }
  }
  else
  {
    # make sure the row is closed.
    if (scalar @{$subRow} == 0)
    {
      $parentRow->{'sub_rows_visible'} = "false";
      $currSubRow->{'visible'} = "false";
    }
    return;
  }
  if (defined $currSubRow->{'sub_rows'})
  {
    foreach my $subCount (keys %{$currSubRow->{'sub_rows'}})
    {
      if (scalar @{$subRow} <= 1 || ($subRow->[1] eq $subCount))
      {
        #${$debugStr} .= "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; expandingSubSubRow = '$subCount'<br>\n";
        my $nextSubRow = $currSubRow->{'sub_rows'}{$subCount};
        $displayChildren = $self->expandSubSubRow($currSubRow, $nextSubRow, $displayZeroRows, $level, $depth + 1, $cellType, $total, $displayChildren, $subRow, $debugStr);
      }
    }
    # reloop over and make sure that the sub_rows_visible flag is correct.
    my $subRowSubRowVisible = "false";
    foreach my $subCount (keys %{$currSubRow->{'sub_rows'}})
    {
      if ($currSubRow->{'sub_rows'}{$subCount}{'visible'} eq "true")
      {
        $subRowSubRowVisible = "true";
      }
    }
    if ($subRowSubRowVisible eq "false")
    {
      $currSubRow->{'sub_rows_visible'} = "false";
    }
  }
}

# displayChildren expandSubSubRow(@data, row_count, displayZeroRows, level, depth, cellType, total, sub_depth1, sub_count1, sub_depth2, sub_count2, displayChildren, subRow, debugStr)
sub expandSubSubRow
{
  my $self = shift;
  my $parentSubRow = shift;
  my $currSubRow = shift;
  my $displayZeroRows = shift;
  my $level = shift;
  my $depth = shift;
  my $cellType = shift;
  my $total = shift;
  my $displayChildren = shift;
  my $subRow = shift;
  my $debugStr = shift;

  #${$debugStr} .= "&nbsp;&nbsp;&nbsp;&nbsp;subrow: $currSubRow->{item_name}";
  #foreach my $name (keys %{$currSubRow})
  #{
  #  ${$debugStr} .= ", $name:$currSubRow->{$name}";
  #}
  #${$debugStr} .= "<br>\n";

  #${$debugStr} .= "&nbsp;&nbsp;&nbsp; depth = '$depth', level = '$level', subRow = [" . join(",", @{$subRow}) . "]<br>\n";

  if (!$self->updateSubRowVisibility($currSubRow, $displayZeroRows, $cellType))
  {
    return $displayChildren;
  }

  my $subTotal = $currSubRow->{'total'};
  if ($level =~ /$negativeOneRE/ || $level >= $depth)
  {
    $parentSubRow->{'sub_rows_visible'} = "true";
    $currSubRow->{'visible'} = "true";
  }
  elsif ($level eq "-2")
  {
    if ($cellType =~ /$tripleAccountsRE/ && $total != 0)
    {
      if (defined $currSubRow->{'sub_rows'})
      {
        $subTotal = $currSubRow->{'total'};
        my @items = keys %{$currSubRow->{'sub_rows'}};
        if ($subTotal != 0)
        {
          if (scalar @items > 1)
          {
            $parentSubRow->{'sub_rows_visible'} = "true";
            $currSubRow->{'visible'} = "true";
            $displayChildren = 1;
          }
          elsif (scalar @items == 1)
          {
            # if the sub account isn't -1, then open it.
            if ($items[0] ne "-1")
            {
              $parentSubRow->{'sub_rows_visible'} = "true";
              $currSubRow->{'visible'} = "true";
              $displayChildren = 1;
            }
            else
            {
              # don't open this sub_row
              $currSubRow->{'sub_rows_visible'} = "false";
            }
          }
        }
        else
        {
          # make sure we don't display this entry
          $currSubRow->{'sub_rows_visible'} = "false";
        }
      }
      else
      {
        my @items = keys %{$parentSubRow->{'sub_rows'}};
        if ($total != 0)
        {
          if (scalar @items > 1)
          {
            $parentSubRow->{'sub_rows_visible'} = "true";
            $currSubRow->{'visible'} = "true";
            $displayChildren = 1;
          }
          elsif (scalar @items == 1)
          {
            # if the sub account isn't -1, then open it.
            if ($items[0] ne "-1")
            {
              $parentSubRow->{'sub_rows_visible'} = "true";
              $currSubRow->{'visible'} = "true";
              $displayChildren = 1;
            }
            else
            {
              $currSubRow->{'sub_rows_visible'} = "false";
            }
          }
        }
        else
        {
          $currSubRow->{'sub_rows_visible'} = "false";
          $currSubRow->{'visible'} = "false";
        }
      }
    }
  }
  elsif ($level eq "-3")
  {
    $parentSubRow->{sub_rows_visible} = "true";
    $currSubRow->{visible} = "true";
    if (scalar @{$subRow} >= 2)
    {
      $currSubRow->{sub_rows_visible} = "true";
    }
    else
    {
      return $displayChildren;
    }
  }
  elsif ($level eq "-4")
  {
    if ($cellType =~ /$tripleAccountsRE/)
    {
      my $subTotal = $currSubRow->{'total'};
      if ($subTotal == 0)
      {
        if (!$displayZeroRows)
        {
          $currSubRow->{visible} = "false";
        }
        return $displayChildren;
      }
    }
  }
  else
  {
    # make sure the sub-row is closed
    if (scalar @{$subRow} == 0)
    {
      $parentSubRow->{'sub_rows_visible'} = "false";
      $currSubRow->{'visible'} = "false";
    }
    return $displayChildren;
  }
  if (defined $currSubRow->{'sub_rows'})
  {
    foreach my $subCount (keys %{$currSubRow->{'sub_rows'}})
    {
      if (scalar @{$subRow} <= 2 || ($subRow->[2] eq $subCount))
      {
        #${$debugStr} .= "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; expandingSubSubSubRow = '$subCount'<br>\n";
        my $nextSubRow = $currSubRow->{'sub_rows'}{$subCount};
        $displayChildren = $self->expandSubSubSubRow($currSubRow, $nextSubRow, $displayZeroRows, $level, $depth + 1, $cellType, $displayChildren, $debugStr);
      }
    }
    # reloop over and make sure the sub_rows_visible flag is correct.
    my $subRowSubRowSubRowVisible = "false";
    foreach my $subCount (keys %{$currSubRow->{'sub_rows'}})
    {
      if ($currSubRow->{'sub_rows'}{$subCount}{'visible'} eq "true")
      {
        $subRowSubRowSubRowVisible = "true";
      }
    }
    if ($subRowSubRowSubRowVisible eq "false")
    {
      $currSubRow->{'sub_rows_visible'} = "false";
    }
  }

  return $displayChildren;
}

# void $self->expandSubSubSubRow(parentSubRow, currSubRow, displayZeroRows, level, depth, cellType, displayChildren)
sub expandSubSubSubRow
{
  my $self = shift;
  my $parentSubRow = shift;
  my $currSubRow = shift;
  my $displayZeroRows = shift;
  my $level = shift;
  my $depth = shift;
  my $cellType = shift;
  my $displayChildren = shift;
  my $debugStr = shift;

  if (!$self->updateSubRowVisibility($currSubRow, $displayZeroRows, $cellType))
  {
    return;
  }

  if ($level =~ /^(-1|-3|-4)$/ || $level >= $depth)
  {
    $parentSubRow->{'sub_rows_visible'} = "true";
    $currSubRow->{'visible'} = "true";
  }
  elsif ($level eq "-2")
  {
    if ($displayChildren)
    {
      $parentSubRow->{'sub_rows_visible'} = "true";
      $currSubRow->{'visible'} = "true";
    }
    else
    {
      $currSubRow->{'visible'} = "false";
    }
  }
  else
  {
    #${$debugStr} .= "sub-sub-row: $currSubRow->{item_name}, depth = '$depth', level = '$level' reached!<br>\n";
    # make sure the sub-sub-row is closed
    $parentSubRow->{'sub_rows_visible'} = "false";
    $currSubRow->{'visible'} = "false";
  }
}

sub format_value
{
  my $self = shift;
  my %args = ( negative_format => 0, sign => "+", value => "0.00", code => 1, @_ );
  my $negative_format = $args{'negative_format'};
  my $value = $args{'value'};
  my $sign = $args{'sign'};
  my $code = $args{code};
  my $result = "";
  my $padding = "1.5em";

  if ($sign ne "+" && $sign ne "-")
  {
    $sign = "+";
  }

  $result = '<span class="number">' if ($code);
  if ($sign eq "-")
  {
    if ($negative_format == 0)
    {
      $result .= '<span style="color: red;">(' . $value . ')</span>';
      $padding = "1em";
    }
    elsif ($negative_format == 1)
    {
      $result .= '<span style="color: red;">-</span>' . $value;
    }
    elsif ($negative_format == 2)
    {
      $result .= $value . '<span style="color: red;">-</span>';
      $padding = "1em";
    }
  }
  else
  {
    $result .= $value;
  }

  $result .= "</span>" if ($code);

  return '<span style="padding-right: ' . $padding . ';">' . $result . "</span>";
}

sub formatUnderlineValue
{
  my $self = shift;
  my %args = ( negative_format => 0, sign => "+", value => "0.00", underlineLength => 0, @_ );
  my $negative_format = $args{'negative_format'};
  my $value = $args{'value'};
  my $sign = $args{'sign'};
  my $underlineLength = $args{underlineLength};

  my $result = "";
  my $padding = "1.5em";
  my $spaces = 0;

  if ($sign ne "+" && $sign ne "-")
  {
    $sign = "+";
  }

  if ($sign eq "-")
  {
    if ($negative_format == 0)
    {
      $spaces = 1;
    }
    elsif ($negative_format == 1)
    {
      $spaces = 2;
    }
    elsif ($negative_format == 2)
    {
      $spaces = 1;
    }
  }
  else
  {
    $spaces = 2;
  }
  my $tmpOut = $self->format_value(%args, code => 0);
  if ($underlineLength)
  {
    my $width = $underlineLength * 7;  # tweak the 7 to find the best pixel ratio.
    my $rightMargin = '1.5em';  # 1.5em cuases the underline to line up on the right with the text below the line.  #3 em causes the right side of the line to line up with almost the right side of the text above the line #1.5 is preferred so that non indented rows dont have underlines pushed over 
    $result = qq{<hr style="float: right; width: ${underlineLength}em; padding: 0; border: none; border-bottom: thin solid black; margin: 0; margin-bottom: 1px; margin-right: $rightMargin;" /><div style="clear: both;">} . $tmpOut ."</div>";
  }
  else
  {
    $result = $tmpOut;
  }

  return $result;
}

sub format_account
{
  my $self = shift;
  my %args = ( hide_number => 0, account_format => 0, name => "", code => "", @_ );
  my $hide_number = $args{'hide_number'};
  my $account_format = $args{'account_format'};
  my $name = $args{'name'};
  my $code = $args{'code'};
  my $output = "";

  if ($hide_number)
  {
    if ($name ne "")
    {
      $output .= $name;
    }
    else
    {
      $output .= "<i>None</i>";
    }
  }
  else
  {
    if ($account_format == 0)
    {
      if ($name ne "")
      {
        $output .= $name;
      }
      else
      {
        $output .= "<i>None</i>";
      }
      $output .= "&nbsp;-&nbsp;";
      if ($code ne "")
      {
        if ($code eq "-1")
        {
          $output .= "(";
        }
        $output .= $code;
        if ($code eq "-1")
        {
          $output .= ")";
        }
      }
      else
      {
        $output .= "<i>None</i>";
      }
    }
    elsif ($account_format == 1)
    {
      if ($code ne "")
      {
        if ($code eq "-1")
        {
          $output .= "(";
        }
        $output .= $code;
        if ($code eq "-1")
        {
          $output .= ")";
        }
      }
      else
      {
        $output .= "<i>None</i>";
      }
      $output .= "&nbsp;-&nbsp;";
      if ($name ne "")
      {
        $output .= $name;
      }
      else
      {
        $output .= "<i>None</i>";
      }
    }
    elsif ($account_format == 2)
    {
      if ($name ne "")
      {
        $output .= $name;
      }
      else
      {
        $output .= "<i>None</i>";
      }
    }
    elsif ($account_format == 3)
    {
      if ($code ne "")
      {
        if ($code eq "-1")
        {
          $output .= "(";
        }
        $output .= $code;
        if ($code eq "-1")
        {
          $output .= ")";
        }
      }
      else
      {
        $output .= "<i>None</i>";
      }
    }
  }

  return $output;
}

1;
__END__;