# Reports.pm - The Object Class that provides a Reports Object
#$Id: Main.pm,v 1.18.4.1 2007/03/27 05:08:23 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.
# Created by Xperience, Inc.

package Portal::Accounting::Reports::Main;
use strict;
use Portal::Accounting::BaseL;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.18.4.1 $';

# global variables.

# modify to represent the commands this state provides.
my %commands = (
  display => "Reports",
);

=head1 NAME

Reports

=head1 SYNOPSIS

  use Portal::Accounting::Reports::Main;
  my $obj = Portal::Accounting::Reports::Main->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

This module is used as the main interface to reports subsystem.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(codeVersion, release, DBVersion)

 Creates a new instance of the Portal::Accounting::Reports::Main module.
 See Portal::Accounting::Base(3) for a listing of required arguments.

 Instantiates an instance of the Reports State. !!Anything passed in here should be
 data that never changes from one run of a command to another such as database handles
 portalSession? input? cookieObj? dont fit that criterion

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{account_formats} = { "0" => "Name - Code", "1" => "Code - Name", "2" => "Name", "3" => "Code" };
  $self->{min_underline_length} =  4;
  $self->{max_underline_length} = 10;
  $self->{default_underline_length} = 4;

  # do validation
  if (!$self->Portal::Accounting::Reports::Main::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  $self->{commands} = \%commands;

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
#   if ($self->numInvalid() > 0 ,, $self->numMissing() > 0)
#   {
#     $self->postfixError($self->genErrorString("all"));
#     return 0;
#   }
  return 1;
}

=item HTMLObject display()

    Implements the default screen for this state.

=cut

sub c_display
{
  my $self = shift;
  my %args = ( command => "display", message => "", @_ );
  my $command = $args{command};
  my $l = $self->{langObj};
  my $message = $args{message};
  my $doc = $self->prepDoc;
  my $url = $self->{baseURL};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $r = $doc->encodeString('Reports::Main');
  my $e = $doc->encodeString('Reports::Editor');
  my $c = $doc->encodeString('Reports::Customize');
  my $d = $doc->encodeString('Reports::Detailed');
  my $t = $doc->encodeString('Reports::Trial');

  my $sql = "SELECT name, id, report_type AS type FROM report_type_tb ORDER BY name";
  my $sth = $self->dbHandle->read($sql);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return undef; }
  my %options;
  while (my $info = $sth->fetchrow_hashref)
  {
    my $which = 'custom';
    if ($info->{id} < 500) { $which = 'system'; }
    $options{$info->{type}}{$which} .= '  <option name="'.$info->{name}.'" value="'.$info->{id}.'">'.$info->{name}."</option>\n";
  }
  my $reports = <<"ENDOF";
<div style="float: right; margin-bottom: 1em;">
  <b>#LANG=General Ledger# #LANG=Reports#</b><br>
  <a href="$url?app=$self->{app}&state=$c&ledgerCode=$ledgerCode&command=display&id=12">#LANG=Profit and Loss#</a><br>
  <a href="$url?app=$self->{app}&state=$c&ledgerCode=$ledgerCode&command=display&id=14">#LANG=Balance Sheet#</a><br>
  <a href="$url?app=$self->{app}&state=$c&ledgerCode=$ledgerCode&command=display&id=6">#LANG=Cash Flow#</a><br>
  <a href="$url?app=$self->{app}&state=$t&ledgerCode=$ledgerCode&command=display&id=6">#LANG=Trial Balance#</a><br>
  <br><br>
  <form name="system" method="post">
    <input type="submit" value="Run System Report" onclick="return runReport(this.form);"/>
    <select name="id">
$options{r}{system}
    </select>
    <input type="hidden" name="app" value="$self->{app}" />
    <input type="hidden" name="state" value="Reports::Customize" />
    <input type="hidden" name="command" value="display" />
    <input type="hidden" name="ledgerCode" value="$ledgerCode" />
    <input type="hidden" name="windowName" value="" />
    <input type="checkbox" name="newWindow" checked="1" /> New Window
  </form>
  <form name="custom" method="post">
    <input type="submit" value="Run Custom Report" onclick="return runReport(this.form);"/>
    <select name="id">
$options{r}{custom}
    </select>
    <input type="hidden" name="app" value="$self->{app}" />
    <input type="hidden" name="state" value="Reports::Customize" />
    <input type="hidden" name="command" value="display" />
    <input type="hidden" name="ledgerCode" value="$ledgerCode" />
    <input type="hidden" name="windowName" value="" />
    <input type="checkbox" name="newWindow" checked="1" /> New Window
  </form>
</div>
ENDOF
  my $trans = <<"ENDOF";
<div>
  <b>#LANG=Transaction Reports#</b><br>
  <a href="$url?app=$self->{app}&state=$d&ledgerCode=$ledgerCode&command=display">#LANG=Search#</a><br>
</div>
ENDOF
  my $editor = <<"ENDOF";
<div style="padding-top: 5em;">
  <b><a href="$url?app=$self->{app}&state=$e&ledgerCode=$ledgerCode&command=display">#LANG=Report Editor#</a></b>
</div>
ENDOF

  $doc->print($self->langSubstitute($reports));
  $doc->print($self->langSubstitute($trans));
  $doc->print($self->langSubstitute($editor));
  $doc->setFocus('javascript');
  $doc->print(<<"FOO");
function runReport(myForm)
{
  var name = getTimeStamp();
  if (myForm.newWindow.checked)
  {
    window.open('', name, 'toolbar=no,location=no,menubar=no,directories=no,copyhistory=no')
    myForm.target = name;
    myForm.windowName.value = name;
  }
  else
  {
    myForm.target = "";
    myForm.windowName.value = '';
  }
  return true;
}
FOO
  return $doc;
}

sub prepDoc
{
  my $self = shift;
  my $doc = $self->SUPER::prepDoc(@_);
#  $doc->setJavascriptInclude(file => $self->{jsDir}."Reports.js");
  return $doc;
}

# log_temp_file - makes the entry in the database that denotes a temp file that needs to be cleaned up later via cron.
# takes  : dbh, filename
# returns: nothing
sub log_temp_file
{
  my $self = shift;
  my %args = ( @_, );
  my $dbh = $self->dbHandle;
  my $sql = "INSERT INTO temp_file_tb (timeof, fname) VALUES (now(), ?)";
  my $file_name = $args{'file_name'};
  if (length $file_name == 0)
  {
    $self->error("Error: log_temp_file.  The filename is required at line " . __LINE__);  return undef;
  }

  my $sth = $dbh->read(sql =>  $sql, plug => [$file_name]);
  if ($dbh->error) { $self->error($dbh->errorMessage); return undef; }
  $dbh->commit;
}

# Takes  : id of report
# Returns: array of arrays where each sub-array is the data for a row in the specified report.
sub GetReportRows
{
  my $self = shift;
  my %args = ( @_ );
  my $dbh = $self->dbHandle;
  my $id = $args{'id'};
  my $sql = "SELECT index, cell_text, cell_type, item_name, '', '', display_level, initial_state, negative_format, other_attributes FROM report_data_tb WHERE id = ? ORDER BY index";
  if (length $id == 0)
  {
    $self->error("Error: GetReportRows. The Report ID is required at line " . __LINE__);  return undef;
  }

  my $sth = $dbh->read(sql => $sql, plug => [$id]);
  if ($dbh->error) { $self->error($dbh->errorMessage); return undef; }
  my $ref = $sth->fetchall_arrayref;
  my @info = @{$ref};

  # now walk the array and make hashes for each row that detail the items and their values.
  my @temp_array = ();
  for (my $i=0; $i < scalar @info; $i++)
  {
    $temp_array[$i] = ();  # make an empty array placeholder.
    my %temp_hash = {};
    $temp_hash{'index'} = $info[$i][0];
    $temp_hash{'cell_text'} = $info[$i][1];
    $temp_hash{'cell_type'} = $info[$i][2];
    $temp_hash{'item_name'} = $info[$i][3];
    $temp_hash{'display_level'} = $info[$i][6];
    $temp_hash{'initial_state'} = $info[$i][7];
    $temp_hash{'negative_format'} = $info[$i][8];
    my $other_attributes = $info[$i][9];
    $temp_hash{'other_attributes'} = $other_attributes;

    my $cell_type = $temp_hash{'cell_type'};
    my @other_attributes = split /\|/, $other_attributes;
    foreach my $item (@other_attributes)
    {
      (my $name, my $value) = split /=/, $item, 2;
      $temp_hash{$name} = $value;
    }
    if ($cell_type eq "text" || $cell_type eq "line" || $cell_type eq "detail")
    {
      $temp_hash{'cell_formula'} = "";
      $temp_hash{'sub_levels'} = "-1";
      $temp_hash{'drillable'} = 0;
      if ($cell_type ne "detail")
      {
        $temp_hash{'activity'} = 1;  # activity
      }
      else
      {
        $temp_hash{'activity'} = 0;  # balances
      }
      $temp_hash{'query_type'} = 0;  # everything
      $temp_hash{'show'} = 1;  # show this row.
      $temp_hash{'underline'} = 0; # don't need to underline the previous row.
    }
    elsif ($cell_type eq "total" || $cell_type eq "compare")
    {
      $temp_hash{'sub_levels'} = "-1";
      $temp_hash{'drillable'} = 0;
      $temp_hash{'activity'} = 1;  # activity
      $temp_hash{'query_type'} = 0;  # everything
      if (!exists $temp_hash{'underline'})
      {
        if ($cell_type eq "total")
        {
          $temp_hash{'underline'} = 1;  # Default to underline the previous row.
        }
        elsif ($cell_type eq "compare")
        {
          $temp_hash{'underline'} = 0;  # Default to not underline the previous row.
        }
      }
    }
    else
    {
      $temp_hash{'underline'} = 0;
    }
    if ($temp_hash{underline} ne '1') { $temp_hash{underline} = 0; }
    $temp_array[$i] = { %temp_hash };
  }

  return @temp_array;
}

# Takes  : id of report
# Returns: "" if not found, report name if found
sub DoesReportExist
{
  my $self = shift;
  my %args = ( @_ );
  my $dbh = $self->dbHandle;
  my $id = $args{'id'};
  my $result = "";  # assume it is not found.
  my $sql = "SELECT name FROM report_type_tb WHERE id = ?";

  if (length $id == 0)
  {
    $self->my_die("Error: DoesReportExist.  The Report ID is required at line " . __LINE__);  return undef;
  }

  my $sth = $dbh->read(sql => $sql, plug => [$id]);
  if ($dbh->error) { $self->error($dbh->errorMessage); return undef; }
  my @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    $result = $info[0];
  }
  return $result;
}


=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

PC & Web Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut