#$Id: Trial.pm,v 1.1.2.3 2007/03/28 04:29:46 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.
# Created by Xperience, Inc.

package Portal::Accounting::Reports::Trial;
use strict;
use Portal::Accounting::BaseL;
use Portal::Accounting::Reports::Objects::Trial;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.1.2.3 $';

# global variables.

# modify to represent the commands this state provides.
my %commands = (
  display => "Trial Balance",
  form => "Trial Balance",
  report => "Trial Balance",
);

=head1 NAME

Reports - Object used to build a Reports for accounting package

=head1 SYNOPSIS

  use Portal::Accounting::Reports;
  my $obj = Portal::Accounting::Reports->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

This module contains payroll tools.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(codeVersion, release, DBVersion)

 Creates a new instance of the Portal::Accounting::Reports module.
 See Portal::Accounting::Base(3) for a listing of required arguments.

 Instantiates an instance of the Reports State. !!Anything passed in here should be
 data that never changes from one run of a command to another such as database handles
 portalSession? input? cookieObj? dont fit that criterion

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module

  # do validation
  if (!$self->Portal::Accounting::Reports::Trial::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  $self->{commands} = \%commands;
  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }
  return 1;
}

=item HTMLObject display()

    Implements the default screen for this state.

=cut

sub c_display
{
  my $self = shift;
  return $self->c_form(@_);
}

sub c_report { my $self = shift; return $self->c_form(@_); }
sub c_form
{
  my $self = shift;
  my $doc = $self->prepDoc(@_);
  my $ledgerCode = $self->{input}->{ledgerCode};
  my %args = ( command => $self->{input}->{command}, doc => undef, @_ );
  $self->{OBJECT} = Portal::Accounting::Reports::Objects::Trial->new(ledgerCode => $self->{input}->{ledgerCode}, langObj => $self->{langObj}, dbHandle => $self->{dbHandle});
  my $command = $args{command};
  my $l = $self->{langObj};
  my $foo = $self->{OBJECT};
  if ($self->{OBJECT}->error) { $doc->print($self->{OBJECT}->errorMessage); return $doc;  }
  if ($command eq "report")
  { #validate user input then react
    if ($foo->populate(%{$self->{input}}) && $foo->validateForm(input => $self->{input}))
    {
      my $je = $foo->select();
      if ($foo->error) { $doc->print($foo->errorMessage); }
      else
      {
        $doc->print($foo->printReport);
        if ($foo->error) { $doc->print($foo->errorMessage); }
      }
      return $doc;
    }
  }
  #display form
  my %form = $foo->printForm();
  if ($foo->error)
  {
    $self->error($foo->errorMessage);
  }
  $doc->print($foo->{formObj}->genErrorString('all'));
  if (defined %form)
  {
    $doc->print(%form);
  }
  else
  {
    $doc->print('error! form is undefined<br>');
    $doc->print("\nformObj:".$foo->{formObj}->error);
  }

  $doc->print($foo->debugString);
  return $doc;
}

sub prepDoc
{
  my $self = shift;
  my $doc = $self->SUPER::prepDoc(@_);
  if (defined $doc) {
    $doc->setJavascriptInclude(file => $self->{jsDir}."Reports/Trial.js");
    $doc->setOnload("init();");
  }
  return $doc;
}
=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

PC & Web Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut